/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */

package osmq.agents;

import java.util.Vector;

import osmq.util.ExceptionListener;
import osmq.net.RemoteService;
import osmq.net.SocketStreamFilter;
import osmq.records.DataRecord;
import osmq.messages.Message;
import java.io.*;

/**
  * A class used by client applications to write messages to a remote peer
  * service application.
  * <p>The BaseNetWriter dynamically locates and connects to
  * the remote service (reader).  Once connected, the BaseNetWriter supports
  * sending messages in String, byte array or DataRecord formats to the remote
  * service's queue.
  */
class BaseNetWriter  implements ExceptionListener
   {
   private String             _servicename = "";
   private int                _portid      = 0;
   private String             _hostname    = "";
   private RemoteService      _rs;
   private ExceptionListener  _el = null;
   private SocketStreamFilter _connection = null;
   private transient boolean  _autocommit = true;
   private long               _physwrites = 0;
   private long               _commitments = 0;

   // holder of messages in native Message format
   private Vector             _messagecache = new Vector();

   public BaseNetWriter()
      {
      _rs = new RemoteService();
      }

   public BaseNetWriter(RemoteService rs)
      {
      _rs = rs;
      }

   /**
     * Sets the ExceptionListener instance that will be
     * called in the event of a network error.
     * @see #getExceptionListener()
     */
   protected void setExceptionListener(ExceptionListener listener)
      {
      _el = listener;
      }

   /**
     * Returns a reference to the ExceptionListener that will be
     * called in the event of a network error.
     * @see #setExceptionListener(ExceptionListener)
     */
   protected ExceptionListener getExceptionListener()
      {
      return _el;
      }

    /**
      * Sets the logical name of the remote service to which messages
      * will be published.  Setting the service name utilizes dynamic
      * discovery to attach to the remote service when the connect()
      * method is called.
      * @see #getRemoteServiceName()
      */
   protected void setRemoteServiceName(String serviceid)
      {
      errorIfConnected();
      _servicename = serviceid;
      if(_servicename != null && _servicename.length() > 0)
         {
         _portid      = 0;
         _hostname    = "";
         }
      }

    /**
      * Returns the logical name of the remote service.
      * @see #setRemoteServiceName(String)
      */
   protected String getRemoteServiceName()
      {
      return _servicename;
      }

    /**
      * Sets the listening port id of the remote service.
      * Setting the Service Address disables dynamic discovery
      * @see #getRemotePortID()
      */
   protected void setRemotePortID(int port)
      {
      errorIfConnected();
      _portid = port;
      if(_portid != 0)
         _servicename = null;
      }

    /**
      * Returns the listening port number of the remote service.
      * If dynamic discovery is used, this will return a zero value.
      * @see #setRemotePortID(int)
      */

   protected int getRemotePortID()
      {
      return _portid;
      }

    /**
      * Sets the ip address / host name of the remote service.
      * Setting the Service Address disables dynamic discovery
      * @see #getRemoteServiceAddress()
      */
   protected synchronized void setRemoteHostName(String address)
      {
      errorIfConnected();
      _hostname = address;
      if(_hostname != null)
         _servicename = null;
      }

    /**
      * Returns the ip address / host name of the remote service.
      * If dynamic discovery is used, this will return a null value.
      * @see #setRemoteServiceAddress(String)
      */

   public String getRemoteHostName()
      {
      return _hostname;
      }

   private void errorIfConnected()
      {
      if(isConnected())
         throw new IllegalStateException("Connection is open");
      }

    /**
      * Returns <i>true</i> if the writer connection has not been closed.
      * @see #disconnect()
      */
   public boolean isConnected()
      {
      return _connection != null;
      }

   private SocketStreamFilter getConnection() throws IOException
      {
      if(!isConnected())
         {
         synchronized(this)
            {
            if(_connection == null)
               {
               _connection = new SocketStreamFilter(_rs, this);
               }
           }
         }
      return _connection;
      }

   public void onException(Exception e)
      {
      if(_el != null)
         _el.onException(e);
      disconnect();
      }

    /**
      * Closes the writer's connection to the remote service.
      * Note that once the connection has been closed, any subsequent
      * calls to publish(String) or publish(byte[]) will result in
      * an IOException being thrown.
      */
   public void disconnect()
      {
      synchronized(this)
         {
         if(isConnected())
            _connection.close();
         _connection = null;
         }
      }

   /**
     * Turns on or off the autocommitment of messages to the remote server.
     * <p> When autocommit is <i>ON</i>, each message that is written gets transmitted
     * to the remote server.
     * <p> When autocommit is <i>OFF</i>, the user must call <code>commit()</code>
     * to send written messages to the server.  Written messages that have not
     * been committed can be discarded by a call to the method rollback().
     * <p> The commit() method transmits all messages
     * that were written since the last call to commit() or rollback().
     * <p><b>NOTE:</b> The default mode is autocommit <i>ON</i>.
     * @throws IOException an attempt to set autocommit to <i>true</i> when
     *         there are messages that have been published but not commited
     *         or rolled back.
     * @see #commit()
     * @see #rollback()
     */
    protected synchronized void setAutoCommit(boolean autocommit) throws IOException
       {
       if(_messagecache.size() > 0 && autocommit == false)
              throw new IOException("Cannot turn on autocommit when there are" +
                                    " published and uncommited messages");
       _autocommit = autocommit;
       }


    /**
      * Used to determine whether auto commit is enabled.
      * @see #setAutoCommit(boolean)
      */
    protected boolean isAutoCommit()
       {
       return _autocommit;
       }

   /**
     * Returns the number of messages that have been published and committed.
     */
    public long getCommittedCount()
       {
       return _physwrites;
       }

   /**
     * Returns the total number of messages that have been published since
     * the connection was last opened or re-opened.
     */
    public long getConnectionPublishedCount()
       {
       return (isConnected()) ?
          _connection.getWrittenCount() : 0;
       }

   /**
     * Returns the number of uncommitted messages currently in the queue.
     * There will be no messages in the queue if autocommit is <i>ON</i>.
     */
    public int getUncommittedCount()
       {return _messagecache.size();}

   /**
     * <p> Discards all messages that were published since the last
     * call to <code>commit()</code> or <code>rollback()</code>.
     * This method is available when autocommit mode is <i>OFF</i>.
     *
     * @throws IOException autocommit mode is <i>ON</i>
     * @see #commit()
     * @see #setAutoCommit(boolean)
     */
   protected synchronized void rollback() throws IOException
       {
       if(_autocommit)
             throw new IOException("Publisher in autocommit mode");
       _messagecache.removeAllElements();
       }

   /**
     * <p> Transmits all messages that were published since the last
     * call to <code>commit()</code> or <code>rollback()</code>.
     * This method is available when autocommit mode is <i>OFF</i>.
     *
     * @throws IOException autocommit mode is <i>ON</i>
     * @see #rollback()
     * @see #setAutoCommit(boolean)
     */
   protected synchronized void commit() throws IOException//, MessageException
      {
      Object  _sv;
      if(_autocommit)
         throw new IOException("Attempt to commit message in autocommit mode");

      while(_messagecache.size() > 0)
         {
         _sv = _messagecache.elementAt(0);
         if(_sv != null)
            _physWrite(_sv);
         _messagecache.removeElementAt(0);
         }
      _commitments++;
      }

   protected synchronized void writeObject(Object o) throws IOException
      {
      // auto commit mode writes each message as it's published
      if(isAutoCommit())
         _physWrite(o);
      else
         _messagecache.addElement(o);
      }

   // this method is overridden in NetMessageWriter
   protected void _physWrite(Object b) throws IOException
       {
       getConnection().writeBytesCounted((byte []) b);
       _physwrites++;
       }

   /**
     * Returns the number of explict executions of the commit()
     * function.
     */
   public long getCommitTransactionCount()
      {return _commitments;}

   /**
     * Opens the publisher's connection to the remote service.
     * Prior to calling this function, either
     * <p> 1) the service name must have been set (for dynamic discovery), or
     * <p> 2) the address and port number of the remote
     * service must have been set.
     * <p>The connection must be opened before publishing of messages.
     */
   public synchronized void connect() throws IOException
       {
       if(!isConnected())
          {
          if(_servicename.length() > 0)
                _rs.setServiceID(_servicename);
          else
             {
             _rs.setPortID(_portid);
             _rs.setHostName(_hostname);
             }
          getConnection();
          }
       }
   } 