/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */

package osmq.agents;

import java.io.IOException;
import java.util.Vector;

import osmq.util.ExceptionListener;
import osmq.net.RemoteService;
import osmq.net.SocketStreamFilter;
import osmq.records.DataRecord;
import osmq.messages.Message;
import osmq.messages.MessageException;

/**
  * A class used by client applications to write messages to a remote peer
  * service application.
  * <p>The NetWriter dynamically locates and connects to
  * the remote service (reader).  Once connected, the NetWriter supports
  * sending messages in String, byte array or DataRecord formats to the remote
  * service's queue.
  */
public class NetMessageWriter  extends BaseNetWriter
   {
   private String _originator;
   private int _trxid;

   public NetMessageWriter()
      {
      super();
      }

   public NetMessageWriter(RemoteService rs)
      {
      super(rs);
      }

   public void setOriginator(String originator)
      {
      _originator = originator;
      }

   public String getOriginator(){return _originator;}

   public void write(Message m) throws IOException, MessageException
       {
       if(m != null)
          {
          // assign default values
          // if no originator in the message, uses the client id  as a default
          if(!m.hasAssignedOriginator() && _originator != null)
              m.setOriginator(_originator);

          // assign a unique message number
          //m.setSequenceNumber(_physwrites + 1);

          // places a time stamp for when the message was generated
          m.setTimeStamp(System.currentTimeMillis());

          // set a corrleation number if this is not an autocommited transaction
          if(!isAutoCommit())
            m.setCorrelationNumber(getCommitTransactionCount() + 1);

          // auto commit mode writes each message as it's published
          writeObject(m);
          }
       else
          System.out.println("Attempt to write null String in ClientPublisherBean");
       }

   // this method is overrides method in BaseNetWriter
   protected void _physWrite(Object b) throws IOException
       {
       Message m = (Message) b;
       m.setSequenceNumber(getCommittedCount() + 1);
       super._physWrite(m.getMessageAsBytes());
       }
       /**
      * Sets the logical name of the remote service to which messages
      * will be published.  Setting the service name utilizes dynamic
      * discovery to attach to the remote service when the connect()
      * method is called.
      * @see #getRemoteServiceName()
      */
   public void setRemoteServiceName(String serviceid)
      {
      super.setRemoteServiceName(serviceid);
      }

    /**
      * Returns the logical name of the remote service.
      * @see #setRemoteServiceName(String)
      */
   public String getRemoteServiceName()
      {
      return super.getRemoteServiceName();
      }

   /**
     * Turns on or off the autocommitment of messages to the remote server.
     * <p> When autocommit is <i>ON</i>, each message that is written gets transmitted
     * to the remote server.
     * <p> When autocommit is <i>OFF</i>, the user must call <code>commit()</code>
     * to send written messages to the server.  Written messages that have not
     * been committed can be discarded by a call to the method rollback().
     * <p> The commit() method transmits all messages
     * that were written since the last call to commit() or rollback().
     * <p><b>NOTE:</b> The default mode is autocommit <i>ON</i>.
     * @throws IOException an attempt to set autocommit to <i>true</i> when
     *         there are messages that have been published but not commited
     *         or rolled back.
     * @see #commit()
     * @see #rollback()
     */
    public synchronized void setAutoCommit(boolean autocommit) throws IOException
       {
       super.setAutoCommit(autocommit);
       }

    /**
      * Used to determine whether auto commit is enabled.
      * @see #setAutoCommit(boolean)
      */
    public boolean isAutoCommit()
       {
       return super.isAutoCommit();
       }

    /**
      * Sets the listening port id of the remote service.
      * Setting the Service Address disables dynamic discovery
      * @see #getRemotePortID()
      */
   public void setRemotePortID(int port)
      {
      super.setRemotePortID(port);
      }

    /**
      * Returns the listening port number of the remote service.
      * If dynamic discovery is used, this will return a zero value.
      * @see #setRemotePortID(int)
      */
   public int getRemotePortID()
      {
      return super.getRemotePortID();
      }

    /**
      * Sets the ip address / host name of the remote service.
      * Setting the Service Address disables dynamic discovery
      * @see #getRemoteHostName()
      */
   public synchronized void setRemoteHostName(String address)
      {
      super.setRemoteHostName(address);
      }

    /**
      * Returns the ip address / host name of the remote service.
      * If dynamic discovery is used, this will return a null value.
      * @see #setRemoteHostName(String)
      */
   public String getRemoteHostName()
      {
      return super.getRemoteHostName();
      }

   /**
     * Sets the ExceptionListener instance that will be
     * called in the event of a network error.
     * @see #getExceptionListener()
     */
   public void setExceptionListener(ExceptionListener listener)
      {
      super.setExceptionListener(listener);
      }

   /**
     * Returns a reference to the ExceptionListener that will be
     * called in the event of a network error.
     * @see #setExceptionListener(ExceptionListener)
     */
   public ExceptionListener getExceptionListener()
      {
      return super.getExceptionListener();
      }

   /**
     * <p> Discards all messages that were published since the last
     * call to <code>commit()</code> or <code>rollback()</code>.
     * This method is available when autocommit mode is <i>OFF</i>.
     *
     * @throws IOException autocommit mode is <i>ON</i>
     * @see #commit()
     * @see #setAutoCommit(boolean)
     */
   public synchronized void rollback() throws IOException
       {
       super.rollback();
       }

   /**
     * <p> Transmits all messages that were published since the last
     * call to <code>commit()</code> or <code>rollback()</code>.
     * This method is available when autocommit mode is <i>OFF</i>.
     *
     * @throws IOException autocommit mode is <i>ON</i>
     * @see #rollback()
     * @see #setAutoCommit(boolean)
     */
   public synchronized void commit() throws IOException//, MessageException
      {
      super.commit();
      }
   }