/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */

package osmq.agents;

import java.util.*;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.io.IOException;
import java.text.DecimalFormat;

import osmq.util.*;
import osmq.records.DataRecord;
import osmq.records.DataRecordBean;
import osmq.messages.Message;
import osmq.messages.MessageFactory;
import osmq.messages.MessageException;
import osmq.mque.*;
import osmq.logs.Logs;
import osmq.net.DiscoveryServer;

/**
  * A class that supports acting as a named service, permitting remote
  * processes to dynamically attach and write messages to the service.
  * <p>The NetReader dynamically listens for "where is" requests, based on
  * its service name, and returns its service IP address and port number.
  * Simultaneously, it accepts TPC/IP connections on its service port, and
  * writes all byte array and string messages sent from the remote "writer"
  * to a local queue.  The class provides methods for accessing these
  * messages in fifo order as a byte array, String or DataRecord.
  */
public class NetReader
  {
  final private static boolean DEFAULT_MODE_BINARY = true;
  final private static int DEFAULT_SIZE_THESHOLD = 1000;
  final private static int DEFAULT_MAX_SIZE = 100000;

  private ExceptionListener       _el = null;
  private MQue                    _pipeline;
  private ReaderConnectionManager _processor = null;

  private boolean                 _isopen        = false;
  private boolean                 _paused = false;

  private boolean                 _binarySocketStreams =
                                          DEFAULT_MODE_BINARY;
  private String                  _observiceName = "";
  private int                     _defaultport   = -1;
  private DiscoveryServer         _discoveryserver = null;
  private int                     _discoveryTtl = -1;

  public NetReader()
     {
     _pipeline = new MQue();
     _pipeline.setPagingThreshold(DEFAULT_SIZE_THESHOLD);
     _pipeline.setMaxThreshold(DEFAULT_MAX_SIZE);
     }

   /**
     * Sets an ExceptionListener that will be notified when there is
     * an exception due to an inability to an I/O or communications failure.
     */
   public void setExceptionListener(ExceptionListener listener)
      {
      _el = listener;
      }

   /**
     * Return the ExceptionListener that will be notified when there is
     * an exception due to an I/O or communications failure.
     */
   public ExceptionListener getExceptionListener()
      {
      return _el;
      }

   /**
     * Sets the in-memory size threshold of the message queue.  When the number
     * of messages exceeds <i>size</i>, excess messages are paged to disk.
     * @see #getOverflowSize()
     */
   public void setOverflowSize(int size)
      {
      _pipeline.setPagingThreshold(size);
      }

   /**
     * Returns the in-memory size threshold of the message queue.  When the number
     * of messages exceeds that threshold, excess messages are paged to disk.
     * @see #setOverflowSize(int)
     */
   public int getOverflowSize()
      {
      return _pipeline.getPagingThreshold();
      }

  /**
    * Sets the maximum capacity threshold of the message queue.  When the queue
    * reaches this capacity (element count), subsequent writes throw an exception.
    * A value of 0 indicates no threshold.
    * @see #getMaximumSize()
    */
   public void setMaximumSize(int size)
      {
      _pipeline.setMaxThreshold(size);
      }

   /**
     * Returns the in-memory count threshold of the message queue.  When the number
     * of messages exceeds that threshold, an exception is thrown.
     * @see #setMaximumSize(int)
     */
   public int getMaximumSize()
      {
      return _pipeline.getMaxThreshold();
      }

   public long getProcessedCount()
      {
      return _pipeline.getPullCount();
      }

   /**
     * Returns true if the queue has more unread messages.
     */
  public boolean hasMessages()
      {
      return !_pipeline.isEmpty();
      }

  /**
    *  Returns the number of message remaining in the queue.
    */
  public long getQueueSize()
      {
      return _pipeline.size();
      }

   /**
     * Returns the next message in the queue as a String.
     * This is a blocking read, which waits until there is
     * a string to return.
     */
  public String getNextString() throws IOException
      {
      return _pipeline.readString();
      }

   /**
     * Returns the next message in the queue as a byte array.
     * This is a blocking read, which waits until there is
     * a byte array to return.
     */
  public byte[] getNextBytePacket() throws IOException
      {
      return _pipeline.readBytes();
      }

   /**
     * Returns the next message in the queue as a DataRecord.
     * This is a blocking read, which waits until there is
     * a DataRecord to return.
     */
  public DataRecord getNextDataRecord() throws IOException
      {
      return new DataRecordBean(getNextBytePacket());
      }

   /**
     * Returns the next message in the queue as an immutable Message.
     * This is a blocking read, which waits until there is
     * a Message to return.
     */
  public Message getNextMessage() throws IOException, MessageException
      {
      return MessageFactory.createImmutableMessage(getNextBytePacket());
      }

  /**
    * Returns the port number that is/will be used by the
    * server to accept remote connections.
    */
   public int getPort()
      {return _defaultport;}

  /**
    * Sets the port number that will be used by the
    * server to accept remote connections.
    */
   public void setPort(int port)
      {_defaultport = port;}

   /**
     * Returns true if the queue server has been opened but
     * is in a "paused" state.
     */
   public boolean isPaused()
      {return _paused;}

   /**
     * Returns the unique name being used by this server for
     * dynamic discovery by remote clients.
     */
   public String getServiceName()
      {return _observiceName;}

   /**
     * Sets the unique name that will be used by this server for
     * dynamic discovery by remote clients.
     */
   public void setServiceName(String sn)
      {_observiceName = sn.toUpperCase();}

   /**
     * Returns true if the server is active (opened) and accepting
     * connections and messages.
     */
   public boolean isOpened()
      {return _isopen;}

    /**
     * Activates the server so that it begins accepting
     * connections and messages.
     */
   public void open() throws Exception
      {
      if(!validName(_observiceName))
          throw new IllegalStateException("Service name not set");
      if(_defaultport  < 1)
          throw new IllegalStateException("Service port number not set");
      if(_discoveryserver == null)
         activateDiscoveryServer();
      if(_processor == null)
        startSocketServerSession();

      _isopen = true;
      }

   /**
     * Places the server in a paused or un-paused mode.
     * In a paused mode, the server halts reading new
     * messages from the remote client (although existing message
     * that have not been read from the local queue can still
     * be accessed).
     */

   public void setPaused(boolean v)
      {
      if(_processor != null)
         _processor.setPauseHandlers(v);
      _paused = v;
      }

  public int getConnectionCount()
     {
     return (_processor != null) ? _processor.getConnectionCount() : 0;
     }

  private void activateDiscoveryServer() throws Exception
     {
     _discoveryserver =
        DiscoveryServer.createAndStart(_el,
                                       _observiceName,
                                       _defaultport,
                                       _discoveryTtl,
                                       _binarySocketStreams);
     _discoveryserver.setUserDefinedName("Discovery Server");
     }


   /**
     * Deactivates the server so that it no longer accepts
     * connections and messages. Any existing connections are
     * terminated.
     */
  public void close()
     {
     Threaded.sleepTenths(2);
     _processor.killAllPublishers();

     // stop the discovery server
     System.out.println("Shutting down discovery server");

     try{
         if(_discoveryserver != null)
            _discoveryserver.stop();
         }
     catch(Exception e)
         {
          Logs.err.println("Error closing discovery server " + e.getMessage());
         }
     finally{_discoveryserver = null;}

     // stop accepting messages from publisher clients
     System.out.println("Shutting down processor");
     try{
        if(_processor != null)
           {
           _processor.stop();
           Threaded.sleepTenths(2);
           }
        }
     catch(Exception e)
        {
        System.err.println("Error closing socket server " + e.getMessage());
        }
     finally{_processor = null;}

     System.out.println("Waiting for Sorting Thread to flush the pipeline");

     try{
        Threaded.sleepTenths(2);
        }
     catch(Exception e)
        {
        System.err.println("Error closing message sorter " + e.getMessage());
        }

     }



  // Initializes and starts the server socket processing thread
  private void startSocketServerSession() throws Exception
     {
     // create the server socket thread
     _processor = new ReaderConnectionManager(_pipeline);

     _processor.setBinarySocketMode(_binarySocketStreams);

     // Set the listening port for the socket server
     _processor.setPortID(getPort());

     // the object should notify me if there is an exception
     if(_el != null)
       _processor.setExceptionListener(_el);

     _processor.setPauseHandlers(_paused);

     // start the socket server (Start its listening / accept mode)
     _processor.start();
     }

  private boolean validName(String name)
     {
     return name != null && name.length() > 0;
     }
  }






