//=========================================================================
// Object  : Fixed
// Package : fixedpoint
// Version : 0.5.1
//
// An implementation of fixed point numbers in Java.  Mathematical
// operations such as addition, multiplication and division are
// implemented, as well as helpfull mathematical functions such as abs(),
// max() and sqrt().
//
// Class Fixed is is extended from java.lang.Number and it overrides the
// appropriate functions to intergrate it into that class heirachy.
//
//=========================================================================
//
// Public Interface Overview
//
//   Mathematical Class Constants
//     Fixed ZERO
//     Fixed ONE
//     Fixed E
//     Fixed PI
//
//   Constructors
//     public Fixed();
//     public Fixed( Fixed fixed );
//     public Fixed( int n );
//     public Fixed( float f );
//     public Fixed( double d );
//
//   Mathematical operations
//     public Fixed add( Fixed fixed );
//     public Fixed sub( Fixed fixed );
//     public Fixed mul( Fixed fixed );
//     public Fixed div( Fixed fixed );
//     public static Fixed add( Fixed a, Fixed b );
//     public static Fixed sub( Fixed a, Fixed b );
//     public static Fixed mul( Fixed a, Fixed b );
//     public static Fixed div( Fixed a, Fixed b );
//
//   Comparison operators
//     public boolean lt( Fixed a );
//     public boolean gt( Fixed a );
//     public boolean le( Fixed a );
//     public boolean ge( Fixed a );
//     public boolean ne( Fixed a );
//     public boolean eq( Fixed a );
//     public static boolean lt( Fixed a, Fixed b );
//     public static boolean gt( Fixed a, Fixed b );
//     public static boolean le( Fixed a, Fixed b );
//     public static boolean ge( Fixed a, Fixed b );
//     public static boolean ne( Fixed a, Fixed b );
//     public static boolean eq( Fixed a, Fixed b );
//
//   Mathematical Functions
//     public Fixed neg();
//     public Fixed abs();
//     public Fixed ceil();
//     public Fixed floor();
//     public Fixed round();
//     public Fixed sqrt();
//     public static Fixed neg( Fixed a );
//     public static Fixed abs( Fixed a );
//     public static Fixed ceil( Fixed a );
//     public static Fixed max( Fixed a, Fixed b );
//     public static Fixed min( Fixed a, Fixed b );
//     public static Fixed floor( Fixed a );
//     public static Fixed round( Fixed a );
//     public static Fixed sqrt( Fixed a );
//
//   Extends Number obligations
//     public int intValue();
//     public long longValue();
//     public float floatValue();
//     public double doubleValue();
//
//   Extends Object obligations
//     public String toString();
//
//==========================================================================
// Implementation Notes
//
// . Java does not catch overflows of it primitive intergral data types,
//   and in turn neither does the Fixed class.
//
// . A long (64 bits) was used for internal representaion as it allows for
//   a much larger ranger of number than int (32 bits).
//
// . A Fixed holds 64 bits, of which 16 bits are the mantisa and 32 bits
//   for the integer component.  The 16 bits left over are used during
//   multiplication and division.
//
//==========================================================================
// Version History
//
// No version histroy was maintained for versions prior to 0.4.
//
// Version :	0.4
// Date	   :	19 Apr 1996
// Change  :	Changed from 32 bit int as internal value to 64 bit long.
//		Although this will slow down the computations on 32 bit
//		processors the extra size of the integer part (from 16 to
//		32 bits) was required for my own computations, and for
// 		computations in the fast 3d packages.
//
// Version :	0.4.1
// Date	   :	19 Apr 1996
// Change  :	Fast integer sqrt() algorithm added. subtract() renamed
//              to sub(), and mult() to mul().
//
// Version :    0.5
// Date    :    19 Apr 1996
// Change  :    Added the comparison operators. Wrong version of 0.4.1
//              was released. Opps.
//
// Version :	0.5.1
// Date    :    April 23 1996
// Change  :	Change to the comment style. Constants ZERO and ONE added.
//              Constructor from double added.
//
// Version :	0.5.2
// Date    :	June 1 1996
// Change  :	Added neg() function.
//
// Version :    0.5.3
// Date    :    July 16 1996
// Change  :    Fixed limitation with float contructor.  Problem, and
//              solution reported by Yan Derbenko (root@ugol.srcc.msu.su)
//
//==========================================================================
// To be done
//
//   Mathematical Functions
// 	exp	log	pow
//
//   Other Functions
//	clone	equals
//
//==========================================================================
// Copyright (C) 1996 Paul Matthews.
//
// This file is part of the FixedPoint package.
//
// The FixedPoint package is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as published
// by the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// The FixedPoint package is distributed in the hope that it will be
// useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with the FixedPoint package; see the file COPYING.  If not, write
// to the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The authour of this package can be contacted at
//   snail : 65 Carslake Loop, Gordon, ACT 2906, Australia
//   email : nudge@interline.com.au
//         : Paul.Matthews@dss.gov.au
//
// Note however,
//
// The integer square root algorithm is a modified version of the one
// presented by Christopher J. Musial in the book Graphics Gems II.
//
//==========================================================================

package osmq.datatypes;

import osmq.util.DoubleFormatter;

public class FixedDecimal extends Number
{
  //==================================================================
  // Public Class Constants
  //==================================================================
  public static final FixedDecimal ZERO        = new FixedDecimal( 0 );
  public static final FixedDecimal ONE         = new FixedDecimal( 1 );
  public static final FixedDecimal E           = new FixedDecimal( Math.E  );
  public static final FixedDecimal PI          = new FixedDecimal( Math.PI );

  private DoubleFormatter df = new DoubleFormatter();
  //LongFormatter lf = new LongFormatter();
  //==================================================================
  // Variabes
  //==================================================================
  protected long value;


  //==================================================================
  // Constructors
  //==================================================================
  public FixedDecimal()
  {}

  public FixedDecimal(String s)
     {this(new Double(s).doubleValue());}

  public FixedDecimal( FixedDecimal fixed )
  {
    this.value = fixed.value;
  }

  public FixedDecimal( int n )
    {
    this.value = ((long)n)<<16;
    }

  public FixedDecimal( float f )
    {
    this.value = (long)((double)f*65536.0+0.5);
    }

  public FixedDecimal( double d )
    {
    this.value = (long)(d*65536.0+0.5);
    }


  public void setCurrency(boolean b)
     {
     df.setCurrency(b);
     }

  public boolean isCurrency()
     {
     return df.isCurrency();
     }
  //==================================================================
  // Mathematical operations
  //==================================================================
  public FixedDecimal add( FixedDecimal fixed )
  {
    this.value += fixed.value;
    return this;
  }

  public FixedDecimal sub( FixedDecimal fixed )
  {
    this.value -= fixed.value;
    return this;
  }

  public FixedDecimal mul( FixedDecimal fixed )
  {
    this.value = (this.value * fixed.value)>>16;
    return this;
  }

  public FixedDecimal div( FixedDecimal fixed )
  {
    this.value = (this.value << 16)/fixed.value;
    return this;
  }

  public static FixedDecimal add( FixedDecimal a, FixedDecimal b )
  {
    FixedDecimal result = new FixedDecimal(a);
    return result.add(b);
  }

  public static FixedDecimal sub( FixedDecimal a, FixedDecimal b )
  {
    FixedDecimal result = new FixedDecimal(a);
    return result.sub(b);
  }

  public static FixedDecimal mul( FixedDecimal a, FixedDecimal b )
  {
    FixedDecimal result = new FixedDecimal(a);
    return result.mul(b);
  }

  public static FixedDecimal div( FixedDecimal a, FixedDecimal b )
  {
    FixedDecimal result = new FixedDecimal(a);
    return result.div(b);
  }

  //==================================================================
  // Comparison operators
  //==================================================================
  public boolean lt( FixedDecimal a )
  {
    return (this.value<a.value);
  }

  public boolean gt( FixedDecimal a )
  {
    return (this.value>a.value);
  }

  public boolean le( FixedDecimal a )
  {
    return (this.value<=a.value);
  }

  public boolean ge( FixedDecimal a )
  {
    return (this.value>=a.value);
  }

  public boolean ne( FixedDecimal a )
  {
    return (this.value!=a.value);
  }

  public boolean eq( FixedDecimal a )
  {
    return (this.value==a.value);
  }

  public static boolean lt( FixedDecimal a, FixedDecimal b )
  {
    return (a.value<b.value);
  }

  public static boolean gt( FixedDecimal a, FixedDecimal b )
  {
    return (a.value>b.value);
  }

  public static boolean le( FixedDecimal a, FixedDecimal b )
  {
    return (a.value<=b.value);
  }

  public static boolean ge( FixedDecimal a, FixedDecimal b )
  {
    return (a.value>=b.value);
  }

  public static boolean ne( FixedDecimal a, FixedDecimal b )
  {
    return (a.value!=b.value);
  }

  public static boolean eq( FixedDecimal a, FixedDecimal b )
  {
    return (a.value==b.value);
  }

  //==================================================================
  // Mathematical Fuctions
  //==================================================================
  public FixedDecimal neg()
  {
    this.value = -this.value;
    return this;
  }

  public FixedDecimal abs()
  {
    this.value = Math.abs(this.value);
    return this;
  }

  public FixedDecimal ceil()
  {
    if( (this.value & 0x000000000000FFFFL)!=0 )
    {
      this.value &= 0xFFFFFFFFFFFF0000L;
      this.value += 0x0000000000010000L;
    }
    return this;
  }

  public FixedDecimal floor()
  {
    this.value &= 0xFFFFFFFFFFFF0000L;
    return this;
  }

  public FixedDecimal round()
  {
    this.value += 0x0000000000007FFFL;
    this.value &= 0xFFFFFFFFFFFF0000L;
    return this;
  }

  public FixedDecimal sqrt()
  {
    if( this.value < 0 ) throw new ArithmeticException();

    long next = this.value>>1;
    long current;

    do
    {
      current = next;
      next = (next + this.value/next)>>1;
    }
    while( next < current );
    this.value = current<<8;

    return this;
  }

  public static FixedDecimal neg( FixedDecimal a )
  {
    FixedDecimal result = new FixedDecimal(a);
    return result.neg();
  }

  public static FixedDecimal abs( FixedDecimal a )
  {
    FixedDecimal result = new FixedDecimal(a);
    return result.abs();
  }

  public static FixedDecimal ceil( FixedDecimal a )
  {
    FixedDecimal result = new FixedDecimal( a );
    return result.ceil();
  }

  public static FixedDecimal floor( FixedDecimal a )
  {
    FixedDecimal result = new FixedDecimal(a);
    return result.floor();
  }

  public static FixedDecimal max( FixedDecimal a, FixedDecimal b )
  {
    return a.value>b.value ? a : b;
  }

  public static FixedDecimal min( FixedDecimal a, FixedDecimal b )
  {
    return a.value<b.value ? a : b;
  }

  public static FixedDecimal round( FixedDecimal a )
  {
    FixedDecimal result = new FixedDecimal( a );
    return result.round();
  }

  public static FixedDecimal sqrt( FixedDecimal a )
  {
    FixedDecimal result = new FixedDecimal( a );
    return result.sqrt();
  }


  //==================================================================
  // Class Number obligations
  //==================================================================
  public int intValue()
  {
    return (int)(this.value>>16);
  }

  public long longValue()
  {
    return this.value>>16;
  }

  public float floatValue()
  {
    return ((float)this.value)/65536.0f;
  }

  public double doubleValue()
  {
    return ((double)this.value)/65536.0;
  }

  //==================================================================
  // Class Object obligations
  //==================================================================
  public String toString()
    {
    //df.setDouble(this.doubleValue());
    //return df.getFormattedString();

    return Float.toString(this.floatValue());
    }

}
