/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */

package osmq.datatypes;

import java.text.*;

/**
  * An immutable data type representing Price.
  */
public class Price extends Number
  {
  private static FactorUtil _facutil = new FactorUtil(2);

  final private static int    DEC_PLACES = _facutil.getDecPlaces();
  final private static String FORMAT_MASK = _facutil.getFormatMask();
  final private static int    FACTOR = _facutil.getFactor();
  
  final public static Price ZeroPrice = new Price();

  private static DecimalFormat df = null;

  private long _internal;

  public Price()
     {
     this(0);
     }

  public Price(double value)
     {
     _internal = _toInternal(value);
     }

  public Price(int value)
     {
     _internal = value * FACTOR;
     }

  public Price(String value)
     {
     this(Double.parseDouble(value));
     }

  public double getDouble()
     {return _internal > 0 ? ((double) _internal) / FACTOR : 0;}

  public int getInt()
     {return _internal > 0 ? (int)  _internal / FACTOR : 0;}
     
  public String toString()
     {
     String dollars = "" + (int) (_internal / FACTOR);
     String cents =  "" + (int)(_internal % FACTOR);
     while(cents.length() < DEC_PLACES)
        cents = cents + "0";

     return "" + dollars + "." + cents;
     }
  
  public String getString()
     {return toString();}
     
  public String getFormattedString()
     {return getDecimalFormatter().format(getDouble());}

  public Price plus(double amount)
     {
     Price m = new Price();
     m._internal = _internal + _toInternal(amount);
     return m;
     }

  public Price plus(Price price)
     {
     Price m = new Price();
     m._internal = this._internal + price._internal;
     return m;
     }

  public Price minus(double amount)
     {
     Price m = new Price();
     m._internal = this._internal - _toInternal(amount);
     return m;
     }

  public Price minus(Price price)
     {
     Price m = new Price();
     m._internal = this._internal - price._internal;
     return m;
     }

  public Price times(int value)
     {
     Price m = new Price();
     m._internal = this._internal * value;
     return m;
     }

  public Price dividedBy(int value)
     {
     if(value == 0)
        throw new IllegalArgumentException("Attempt to divide by zero");
     Price m = new Price();
     // divide the two operands and round up the result
     m._internal = (long) (((double)_internal / value) + 0.5);
     return m;
     }

  public boolean greaterThan(Price price)
     {
     return (this._internal > price._internal) ;
     }

   public boolean lessThan(Price price)
     {
     return (this._internal < price._internal) ;
     }

  public boolean notEqual(Price price)
     {
     return (this._internal != price._internal) ;
     }

  public boolean equals(Price price)
     {
     return (this._internal == price._internal) ;
     }

  private static long _toInternal(double amt)
      {
      return (long) Math.floor((amt * FACTOR) + 0.5d);
      }

  // abstract method implementations that are defined
  // in Number class
  public double doubleValue()
     {return getDouble();}
  public float floatValue()
     {return (float) getDouble();}
  public int intValue()
     {return getInt();}
  public long longValue()
     {return _internal > 0 ? _internal / FACTOR : 0;}


      
   private static DecimalFormat getDecimalFormatter()
      {
      if(df == null)
         {
         synchronized(Price.class)
            {
            if(df == null)
              df = new DecimalFormat(FORMAT_MASK);
            }
         }
      return df;
      }

   final private static String templateBuilder(String s, int places)
       {
       while(places-- > 0)
          s = s + "0";
       return s;
       }

   private static FactorUtil getFactorUtil()
      {
      if(_facutil == null)
         {
         synchronized(Currency.class)
            {
            if(_facutil == null)
               _facutil = new FactorUtil(2);
            }
         }
      return _facutil;
      }

   public static void main(String[] args)
      {
      Price p = new Price("10000.00");
      System.out.println(p.getString());
      System.out.println("add 55.34 = " + p.plus(55.34));
      System.out.println("multiply by 10 = " + p.times(10));
      System.out.println("Divide by 10 = " + p.dividedBy(10));
      System.out.println("divided by 5 = " + p.dividedBy(5));
      System.out.println("divided by 3 = " + p.dividedBy(3));
      System.out.println("formatted = " + p.getFormattedString());
      }

   }





