/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */


package osmq.io;

import java.net.*;
import java.io.*;
import java.util.*;
import java.math.*;

import osmq.util.DynByteArray;


/**
  * An input stream filer that supports reading an input stream
  * as packets of bytes.  Each packet is preceeded with a size.
  * <p> Note this class assumes a single reader and is NOT thread safe.
  */
public class PacketInputStream extends FilterInputStream implements ReaderStream
   {
   // large enough to read 4-byte sizes
   private boolean _eof = false;
   private int _readsize = 0;
   private byte[] readbuffer = new byte[Packet.INT_BYTE_ARRAY_SZ];
   private DynByteArray array = new DynByteArray();

   public PacketInputStream(InputStream in)
      {
      super(in);
      }

    /**
      * Returns the next packet of bytes from the underlying
      * InputStream and copies them into the byte array passed by
      * the calling function.
      */
    public  int readBytes(byte[] target)  throws IOException
        {
        if(target == null)
           throw new IllegalArgumentException("target array is null");

        // first read the size portion of the packet
        int sz = readSize();

        // a size less than zero indicates EOF
        if(sz < 1)
           return -1;

        // make certain the passed array is large enough to hold the result
        if(sz > target.length)
           throw new IllegalArgumentException("target size < packet size");

        // read the data bytes into the target array
        return read(target, 0, sz);
        }

    /**
      * Returns true if the end of the stream was encountered.
      */
    public boolean isEOF()
       {return _eof;}

       
    /**
      * Returns the next packet of bytes from the underlying
      * InputStream returns it as a byte array.
      * <p>
      * This method blocks until the next packet is available.
      * <p>
      * @return     a String containing the packet bytes.  The number
      *             of packet bytes in the array can be determined by
      *             calling getLastPacketSize()
      * @exception  IOException  if an I/O error occurs.
      */
    public  String readln() throws IOException
        {
        try{
           // read the size of the next packet
           int rdsz = readSize();
           if(rdsz < 1)
              throw new IOException("Error reading input stream");

           // make certain target is large enough
           array.initializeBuffer(rdsz);

           // fill it with packet data
           if(read(array.getBuffer(), 0, array.getContentSize()) < 1)
              throw new IOException("Error reading input stream");

           // return string version of byte array
           return array.toString();
           }
        catch(EOFException e)
           {
           throw new IOException(e.getMessage());
           }
        }

  public int getLastPacketSize()
     {
     return _readsize;
     }
     
  /**
    * Returns the next packet of bytes from the underlying
    * InputStream into a newly allocated byte array.
    * This method blocks until the next packet is available.
    * <p>
    * @return     a byte array containing the packet.  The number
    *             of packet bytes in the array can be determined by
    *             calling getLastPacketSize()
    * @exception  IOException  if an I/O error occurs.
    */
  public  byte[] readBytes() throws IOException
        {
        try{
           // read the size of the next packet
           int rdsz = readSize();
           byte[] returnpacket;
           if(rdsz < 1)
              throw new IOException("Error reading input stream");

           // allocate the byte array
           returnpacket = new byte[rdsz];

           // fill it with packet data
           if(read(returnpacket, 0, rdsz) < 1)
              throw new IOException("Error reading input stream");
           return returnpacket;
           }

        catch(EOFException e)
           {
           throw new IOException(e.getMessage());
           }
        }

   /**
     * Reads <code>len</code> bytes of data from the input stream into
     * an array of bytes.
     * The number of bytes actually read is returned as an integer.
     * <p>
     * <i>NOTE</i>  This method blocks until input <code>len</code> is available,
     * end of file is detected, or an exception is thrown.
     * This is differs from the normal non-blocking InputStream contract that
     * may return less than <code>len</code>
     *
     * <p> If <code>b</code> is <code>null</code>, a
     * <code>NullPointerException</code> is thrown.
     *
     * <p> If <code>off</code> is negative, or <code>len</code> is negative, or
     * <code>off+len</code> is greater than the length of the array
     * <code>b</code>, then an <code>IndexOutOfBoundsException</code> is
     * thrown.
     *
     * <p> If <code>len</code> is zero, then no bytes are read and
     * <code>0</code> is returned; otherwise, there is an attempt to read at
     * least one byte. If no byte is available because the stream is at end of
     * file, the value <code>-1</code> is returned; otherwise, at least one
     * byte is read and stored into <code>b</code>.
     *
     * <p> The first byte read is stored into element <code>b[off]</code>, the
     * next one into <code>b[off+1]</code>, and so on. The number of bytes read
     * is, at most, equal to <code>len</code>. Let <i>k</i> be the number of
     * bytes actually read; these bytes will be stored in elements
     * <code>b[off]</code> through <code>b[off+</code><i>k</i><code>-1]</code>,
     * leaving elements <code>b[off+</code><i>k</i><code>]</code> through
     * <code>b[off+len-1]</code> unaffected.
     *
     * <p> In every case, elements <code>b[0]</code> through
     * <code>b[off]</code> and elements <code>b[off+len]</code> through
     * <code>b[b.length-1]</code> are unaffected.
     *
     * <p> If the first byte cannot be read for any reason other than end of
     * file, then an <code>IOException</code> is thrown. In particular, an
     * <code>IOException</code> is thrown if the input stream has been closed.
     *
     * <p> The <code>read(b,</code> <code>off,</code> <code>len)</code> method
     * for class <code>InputStream</code> simply calls the method
     * <code>read()</code> repeatedly. If the first such call results in an
     * <code>IOException</code>, that exception is returned from the call to
     * the <code>read(b,</code> <code>off,</code> <code>len)</code> method.  If
     * any subsequent call to <code>read()</code> results in a
     * <code>IOException</code>, the exception is caught and treated as if it
     * were end of file; the bytes read up to that point are stored into
     * <code>b</code> and the number of bytes read before the exception
     * occurred is returned.  Subclasses are encouraged to provide a more
     * efficient implementation of this method.
     *
     * @param      b     the buffer into which the data is read.
     * @param      off   the start offset in array <code>b</code>
     *                   at which the data is written.
     * @param      len   the maximum number of bytes to read.
     * @return     the total number of bytes read into the buffer, or
     *             <code>-1</code> if there is no more data because the end of
     *             the stream has been reached.
     * @exception  IOException  if an I/O error occurs.
     * @see        java.io.InputStream#read(byte[], int, int)
     */
   public int read(byte[] b, int off, int len) throws IOException
       {
       int n = 0;
       while (n < len)
          {
          // read into the array bytes from the stream, for a length
          // not to exceed the balance of len (not already read)
          int count = this.in.read(b, off + n, len - n);
          if (count < 0)
             {
             _eof = true;
             return -1;
             }
          n += count;
          }
      return len;
      }

   /**
     * Reads the next four bytes from the stream and converts it to
     * an interger.  size of the next data packet and returns it as an integer.
     *
     * <p>
     * This method blocks until the next four bytes are available.
     * <p>
     * This method performs the call <code>read(byte[], 0, 4)</code>
     * and and translates the result into an integer.  It will block until
     * the integer size is non-zero.
     *
     * @return     the size of the next logical data packet, or
     *             <code>-1</code> if there is no more data because the end of
     *             the stream has been reached.
     * @exception  IOException  if an I/O error occurs.
     */
   final protected int readSize() throws IOException
       {
       // read the size, but ignore a size of zero
       try{
          _readsize = 0;
          // ignore packet sizes of zero -- this indicates a keep-alive
          // packet was transmitted
          while (_readsize == 0)
            {
            read(this.readbuffer, 0, Packet.INT_BYTE_ARRAY_SZ);
            _readsize = bytesToInt(this.readbuffer);
            }
          return _readsize;
          }
          
       catch(EOFException e)
           {
           _eof = true;
           return -1;
           }
       }

   /**
     * Combine the four bytes into a word, and conform to little endian.
     * @return An integer value for the given four bytes.
     */
   private int bytesToInt(byte[] value) throws IOException
       {
       return (
                ((((int) value[3]) & 0xff) << 24)
              | ((((int) value[2]) & 0xff) << 16)
              | ((((int) value[1]) & 0xff) <<  8)
              | ((((int) value[0] & 0xff))));
       }

   }
