/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */
 
package osmq.io;

import java.net.*;
import java.io.*;
import java.util.*;
import java.math.*;

/*
 * An OutputStream filter that supports writing byte arrays to an
 * output stream, and including the size of the array as part of the
 * data that is written.
 * <p>The corollary PacketInputStream is able to interpret the size
 * of packets on the stream, and returns byte arrays in packets
 * of the size that were written.
 * <p> This class is thread safe.
 */
public class PacketOutputStream extends FilterOutputStream implements WriterStream
   {
   final public static byte[] NULL_INT_ARRAY = new byte[Packet.INT_BYTE_ARRAY_SZ];
   private byte[] writebuffer = new byte[Packet.INT_BYTE_ARRAY_SZ];

   public PacketOutputStream(OutputStream out)
      {
      super(out);
      }

   /**
     * Writes a null object to the stream which is ignored by
     * the corresponding PacketInputStream.   This method is used
     * for testing whether a socket stream is still active.
     */
   public void writeNull() throws IOException
      {
      synchronized (this)
         {
         writeSize(0);
         }
      }

   public void println(String s) throws IOException
      {
      this.write(s.getBytes(), 0, s.length());
      }


   public void writeBytes(byte[] value) throws IOException
      {
      this.write(value, 0, value.length);
      }
      
   /**
     * Writes the byte array to the underlying output stream,
     * maintaining size information so that the corollary BytePacketReader
     * can fetch packets in the same byte size units.
     * <p>This method overrides the base method of any underlying
     * OutputStream implementation.
     */
   public void write(byte[] b, int offset, int size) throws IOException
        {
        synchronized (this)
           {
           // write the length of the data packet as an int
           writeSize(size);

           // then write the data packet
           out.write(b, offset, size);
           }
        }

  
   /**
     *
     * Writes an integer (size) to the output stream as 4 bytes.
     *
     * @param      b   the data.
     * @exception  IOException  if an I/O error occurs. In particular, 
     *             an <code>IOException</code> may be thrown if the 
     *             output stream has been closed
     * @see        osmq.io.PacketInputStream#readSize()
     */
   final protected void writeSize(int v) throws IOException
       {
       if(v > 0)
          {
          out.write(intToBytes(v, writebuffer),
                    0, Packet.INT_BYTE_ARRAY_SZ);
          }
       else
          {
          out.write(NULL_INT_ARRAY, 0, Packet.INT_BYTE_ARRAY_SZ);
          }
       }

    private  byte[] intToBytes(int word, byte[] target) throws IOException
        {
        target[0] = (byte)  (word & 0x000000ff);
        target[1] = (byte) ((word & 0x0000ff00) >> 8) ;
        target[2] = (byte) ((word & 0x00ff0000) >> 16);
        target[3] = (byte) ((word & 0xff000000) >> 24);
        return target;
        }
   }
