/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */

package osmq.mcns;

import java.net.*;
import java.io.*;
import java.util.*;
import java.text.*;

/**
  * Sends a string to a multicast group/port, and waits for a response. The
  * response is a standard datagram UDP message.
  * <p>Unlike the MCResponder, this method does not require use of a
  * MulticastSocket, because it does not join a multicast group.
  * It simply writes a datagram to the multicast channel
  * as if it were a standard IP address/port combination.
  */

public abstract class MCRequestor
  {
  private boolean _enr = true;  // flag to throw exception on no response from responder
  private int     _channel = MCNS.DISCOVERY_CHANNEL;

  protected MCRequestor()
     {
     _channel = getDefaultChannel();
     }

  abstract protected int getDefaultChannel();

  /**
    * Writes a string to the multicast channel and wait for a response.
    * It will try several times if no response is returned, finally returning
    * a null.
    * @return The string representation of data returned by the respondant as
    * a result of the send request.
    */
  final public synchronized  String sendRequest(String value) throws IOException
     {
     byte writebuf[];
     byte readbuf[] = new byte[MCNS.READBUF_SIZE];
    // RemoteService dr;

     // each iteration uses a new socket
     DatagramSocket socket = null;
     DatagramPacket packet;

     if(value == null || value.length() == 0)
        throw new MCNSException("value not set in MCRequestor prior to send()");
     try{

        int ctr = 0; // counter for number of attempts

        // create a datagram socket using any available port
        socket = new DGSocket();

        // timeout for listening for respondant
        socket.setSoTimeout(MCNS.CLIENT_SOCKET_TIMEOUT);
        writebuf = value.getBytes();
        while(ctr++ < MCNS.MAX_LOOKUP_TRIES)
           {
           // Create packet to the channel of the standard multicast group
           packet = new DatagramPacket(writebuf,
                                       writebuf.length,
                                       InetAddress.getByName(NetworkScope.getScope().getGroupAddress()),
                                       getChannel());

           // send the packet, assuming a respondant will be listening to this channel
           socket.send(packet);

           // create a packet to receive the return results,
           // and listen for a response from a respondant
           packet = new DatagramPacket(readbuf, readbuf.length);
           try{
              socket.receive(packet);
              }

           // If timed out waiting for response, loop back to
           // top if timed out, and try again
           catch(InterruptedIOException  e)
              {
              continue;
              }

           // If the returned packet is not empty, return its value
           // Otherwise, loop back and try again
           if(packet != null)
              {
              return new String(packet.getData());
              }
           else
              continue;
           }

        // either timed out or returned packet was empty
        if(_enr)
            {
            throw new MCNSException("No respone to request from remote peer");
            }
        return null;
        }

     catch(IOException e)
        {
        System.err.println("Error in MCRequestor.sendRequest()" + e.getMessage());
        throw e;
        }

     finally
        {
        if(socket != null)
            socket.close();
        socket = null;
        }
     }

  public void setExceptionOnNoResponse(boolean enr)
     {
     _enr = enr;
     }

  public boolean isExceptionOnNoResponse()
     {
     return _enr;
     }

  public void setChannel(int channel)
     {
     _channel = channel;
     }

  public int getChannel()
     {
     return _channel;
     }

  }

