/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */
 

package osmq.messages;

import java.util.*;
import java.math.*;
import java.io.Serializable;

import osmq.records.*;
import osmq.datarows.*;
import osmq.util.Immutable;
import osmq.util.ImmutableException;

/**
  * An entity relational message (ERMessage) contains entity data
  * values, and supports mapping those data values to/from Java
  * data types, and routing them as a unit to remote processes.  
  */
public interface ERMessage extends Message, DataRow, KeyedDataRow, Cloneable, Serializable
   {

   /**
     * Sets the base index used to reference elements in the message body.
     * If not set explicitly, this index defaults to 1.
     * @param value the new base index value
  	 */
   void setOffsetIndex(int value);


   /**
     * Returns the base index currently used to reference elements in the message body.
     * @see #setOffsetIndex(int)
	   */
   int getOffsetIndex();


   /**
     * Returns true if this is an immutable message.  The attributes and values
     * of an immutable message cannot be changed, and attempts to modify them
     * results in an ImmutableException being thrown.
     * @see #setImmutable()
     */
   boolean isImmutable();


   /**
     * Clears the value at the designated position, setting it to null (or zero if numeric)
     * @param posn the first position is 1, the second is 2, ...
     * @exception ImmutableDataException if the message is immutable.
     * @see #clearBody()  
     */
   void clear(int posn);

   /**
     * Returns the value of the designated position in the message body.
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the value as an Object or null  if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a 
	   *             valid position.
     */
   Object getObject(int posn);

   /**
     * Returns the value at the designated position in the message body as a char
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the value as a char or 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */
   char getChar(int posn);

   /**
     * Returns the value at the designated position in the message body as a byte
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the value as a byte or 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */
   byte getByte(int posn);

   /**
     *  Returns the value at the designated position in the message body as a boolean
	 *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the value as a boolean or a null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */
   boolean getBoolean(int posn);


   /**
     *  Returns the value at the designated position in the message body as a float
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the position value as a double or 0.0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
	 *             valid position.
     */
   float getFloat(int posn);


   /**
     * Returns the value at the designated position in the message body as a double
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the position value as a double or 0.0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
	   *             valid position.
     */
   double getDouble(int posn);

  /**
     *  Returns the value at the designated position in the message body as a BigDecimal
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the position value as a BigDecimal or null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     */
   BigDecimal getBigDecimal(int posn);


   // -------------------- POSITIONAL GETTERS ----------------------

   /**
     * Returns the value at the designated position in the message body as an integer.
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the value as an int or a 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */
   int getInt(int posn);


   /**
     *  Returns the value at the designated position in the message body as a long integer
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the position value as a short or 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if columnIndex is not a valid element position.
     */
   long getLong(int posn);


   /**
     *  Returns the value at the designated position in the message body as a short integer
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the position value as a short or 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if columnIndex is not a valid element position.
     */
   short getShort(int posn);


   /**
     * Returns the value of the designated position in the message body as a String
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the position value as a String or a null String if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */
   String getString(int posn);


   /**
     *  Returns the value at the designated position in the message body as a String
	 *  with all characters forced into upper case.
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the value as a String or a null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid element position.
     */
   String getUpperCaseString(int posn);


   /**
     *  Returns the value at the designated position in the message body as a java.util.Date
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the position value as a java.util.Date, or null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if columnIndex is not a valid element position.
     * @see #setDate(int, Date)
     */
   Date getDate(int posn);


   /**
     *  Returns the value at the designated position in the message object as an SQL date
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the position value as a java.sql.Date object, or null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
	 *             valid position.
     * @see #setDateString(int, String)
     */
   java.sql.Date getSQLDate(int posn);

   /**
      * Returns the value at the designated position in the message object
      * as a formatted date string in CCYY-MM-DD format.
      *
      * @param positionIndex the first position is 1, the second is 2, ...
      * @return  the position value as a String or null if the position is empty.
      * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
	  *             valid position.
      * @see #setDateString(int, String)
      */
   String getFormattedDateString(int posn);

   /**
     * Returns the value at the designated position in the message object
     * as an unformatted date string in CCYYMMDD format.  (This is the normal
     * storage format for a date).
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the position value as a String or null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
	 *             valid position.
     * @see #setDateString(int, String)
     */
   String getUnformattedDateString(int posn);

   /**
     * Determines whether a position has a value (versus null)
     *
     * @param posn the position containing the value to be tested
     * @return  <tt>true</tt> of there is a non-null value at position p, else <i>false</i>
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */
   boolean hasAValue(int posn);

   /**
     * Determines if a position has no value (is empty)
     *
     * @param posn the position containing the value to be tested
     * @return  <tt>true</tt> if there is a null value (no value) at position p, else <tt>false</tt>
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */
   boolean isEmpty(int posn);


   /**
     * Sets the value at the designated position in the message body
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if p is not a valid element position.
     * @exception ImmutableException if the message object is immutable.
     */
   void setObject(int posn, Object value);


   /**
     * Sets the element at the designated position in the message body to a String value
     *
     * @param posn the element's relative position (1 to n)
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if p is not a valid element position.
     * @exception ImmutableException if the message object is immutable.
     */
   void setString(int posn, String s);


   void setTrimmedString(int posn, String s);


   /**
     * Sets the value at the designated position in the message body to null
     *
     * @param posn the first position is 1, the second is 2, ...
     * @exception  ArrayIndexOutOfBoundsException if p is not a valid element position.
     * @exception ImmutableException if the message object is immutable.
     */
   void setNull(int posn);


   /**
     * Sets the element at the designated position in the message body to a byte value
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the message is immutable.
     */
   void setByte(int posn, byte value);



   /**
     * Sets the element at the designated position in the message body to a boolean value
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the message is immutable.
     */
   void setBoolean(int posn, boolean value);


   /**
     * Sets the element at the designated position in the message body to a char value
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the message is immutable.
     */
   void setChar(int posn, char value);


   /**
     * Sets the element at the designated position in the message body to a float value
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid  position.
	   * @exception ImmutableException if the message is immutable.
     */
   void setFloat(int posn, float value);


   /**
     * Sets the element at the designated position in the message body to a double value
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid  position.
     * @exception ImmutableException if the message is immutable.
     */
  void setDouble(int posn, double value);

  void setDecimalString(int posn, String value);

   /**
     * Sets the element at the designated position in the message body to a BigDecimal value
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid  position.
     * @exception ImmutableException if the message is immutable.
     */
  void setBigDecimal(int posn, BigDecimal value);


    /**
     * Sets the element at the designated position in the message body to a Date value
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid  position.
	   * @exception ImmutableException if the message is immutable.
     */
  void setDate(int posn, Date value);


   /**
     * Sets the element at the designated position in the message body to an int value
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
	   * @exception ImmutableException if the message is immutable.
     */
  void setInt(int posn, int value);


   /**
     * Sets the element at the designated position in the message body to a long value
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the message is immutable.
     */
  void setLong(int posn, long value);

   /**
     * Sets the element at the designated position in the message body to a short value
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the message is immutable.
     */
  void setShort(int posn, short value);

   /**
     * Sets the element at the designated position in the message body to an
	   * upper-case String value
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the message object is immutable.
     */
  void setUpperCaseString(int posn, String value);


  /**
     * Sets the element at the designated position in the message body to an date value.
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value in CCYYMMDD or CCYY-MM-DD format.  (If the string
     *        contains hyphens, they are removed.)
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
	 * @exception ImmutableException if the message object is immutable.
     */
  void setDateString(int posn, String value);

   /**
     * Clear all the body elements in the ERMessage object to null (or zero if numeric)
     * @exception ImmutableException if the message is immutable.
     */
  void clearBody();


   /**
     * Clear all the header elements in the ERMessage object to null (or zero if numeric)
     * @exception ImmutableException if the message is immutable.
     */
  void clearHeader();

  /**
    * Sets the message state to "immutable."  The values of
    * an immutable message cannot be modified.  Any attempt to
    * modify those values will result in an ImmutableException being
    * thrown.
    * @see osmq.util.ImmutableException
    */
  void setImmutable();

   /**
     * Sets the message header values.
     * @exception ImmutableDataException if the message is immutable.
     */
   void setHeader(MessageHeader header);


   /**
     * Returns the message header values.
     */
   MessageHeader getHeader();


   /**
     * Returns the header data as an array of bytes.
     */
   byte[] getHeaderBytes();


   /**
     * Sets the header data from an array of bytes.
     * @exception ImmutableDataException if the message is immutable.
     */
   void setHeaderBytes(byte[] value);


   /**
     * Returns the client (body) data as an array of bytes.
     */
   byte[] getClientBytes();

      
   /**
     * Sets the client (body) data from values in a message body.
     * Changes made to the DataRecord will change the corresponding
     * values in the message body.
     * @exception ImmutableDataException if the message is immutable.
     */
   void setClientBytes(DataRecord ds);


   /**
     * Sets the client (body) data from an array of bytes.
     * @exception ImmutableDataException if the message is immutable.
     */
   void setClientBytes(byte[] data);


   /**
     * Returns the message header and body as an array of bytes.
     */
   byte[] getMessageAsBytes();


   /**
     * Returns the message body as a delimited String.
     */
   String getElementsAsString();

      
   /** Returns <i>true</i> if the <i>lastfetched</i> element was a null object, or
     * an object with a string length of zero.  (The <i>last fetched</i>
     * is the element referenced by the last <code>getXXX()</code> method call.)
     */
  boolean wasNull();


   /**
     * Returns the topic of the message.
     *
     * @return  A string containing the topic. The topic defines the
     * message contents for a subscriber (number of elements and the
     * data type of each element).
     */
   String getTopic();


   /**
     * Sets the topic that identifies the message contents.
     * @exception ImmutableDataException if the message is immutable.
     */
   void setTopic(String topic);


   /**
     * Returns the unique identifier for the data contained in the message.
     * (This "key" is relative to the the topic.)
     *
     * @return  A string containing the unique identifier.
     */
   String getKey();


   /**
     * Sets the unique identifier for the data contained in the message.
     * (This "key" is relative to the the topic.)
     *
     * @param s the unique key value
     */
   void setKey(String s);


   /**
     * Sets the unique identifier for the data contained in the message.
     * (This "key" is relative to the the topic.)
     *
     * @param n the unique key value
     */
   void setKeyAsInt(int n);


   int getKeyAsInt();


   /**
     * Returns the identifier of the event or activity that caused the
     * transaction to be generated.  This is a value that is based on the topic.
     * <p>
     * Standard events used for down-stream subscribers that maintain
     * databases are ERMessage.TRANSACTION_ADD, ERMessage.TRANSACTION_DELETE
     * and ERMessage.and TRANSACTION_UPDATE.
     *
     * @return  A string containing the transaction event.
     */
   String getTransaction();

   /**
    * Returns the first character of the event identifer, or a null character if
    * there is no event identifer
    */
    char getTransactionAsChar();

  /**
    * Sets the transaction identifer.  This identifier is useful to
    * subscribers that perform database updates based on topical messages.
    * (The identifiers typically translate into add/update/delete functions)
    */
    void setTransaction(String e);

  /**
    * Sets the transaction identifer to a char value.
    * @see #setTransaction(String)
    */
    void setTransactionAsChar(char c);

    /**
      * Gets the identifier used to identify a group of transactions
      * as related.  The relationship is likely a unit-of-work that
      * resulted in tranaction-related topics being generated.
      * This is an optional header attribute.
      *
      * @return  An int containing the correlation identifier.
      */
    long getCorrelationNumber();


    /**
      * Sets the identifier used to identify a group of transactions
      * as related.  The relationship is likely a unit-of-work that
      * resulted in tranaction-related topics being generated.
      * This is an optional header attribute, and is set automatically
      * by the Publisher.
      *
      * @param n the correlation identifier
      */
    void setCorrelationNumber(long n);

    String getCorrelationString();

    /**
      * Sets the identifier used to identify a group of transactions
      * as related.
      * @see #setCorrelationNumber(long)
      *
      * @param s the correlation identifier
      */
    void setCorrelationString(String s);


    /**
      * Returns the identifier of the message sender (publisher).  The
      * sender identifies a unique client process (instance).
      *
      * @return  A string containing the message sender.
      */
    String getOriginator();

    /**
      * Sets the identifier of the message sender (publisher).  The
      * sender identifies a unique client process (instance).
      * <p> Normally, this value is not set by an application, and
      * defaults to the client identifier used to attach to the message broker.
      *
      * @param t the message originator (sender)
      */
    void setOriginator(String t);

    /**
      * Returns the identifier of the message recipient, or
      * null if the message is intended for all subscriber to the
      * message topic.
      *
      * @return  A string containing the recipient identifier.
      */
    String getRecipient();

    /**
      * Sets the unique recipient that will receive the message.
      * If a recipient value is set, only that recipient will receive
      * the message. (Point-to-point messaging).  Setting the recipient
      * should be reserved for clients requesting a service (type of
      * TYPE_REQUEST) and for OLTP systems that receive a request that
      * is being rejected.  In the latter case, the OLTP system would send
      * a point-to-point rejection message to the originator of the request.
      *
      * @param value the intended message recipient
      */
    void setRecipient(String t);

    /**
      * Identifies the message event type -- normally either MessageType.NOTIFICATION
      * or MessageType.REQUEST.
      *
      * @return  A char containing either the value MessageType.NOTIFICATION or
      * MessageType.REQUEST, MessageType.DENIAL, MessageType.BATCH_CONTROL or
      * MessageType.TOPIC_CHG.
      */
    char getType();


    /**
      * Sets the message type. The message type is the category
      * of event that caused a message to be generated.
      * Standard typess include  MessageAttributes.TYPE_NOTIFICATION
      * and MessageAttributes.TYPE_REQUEST.
      * <p>TYPE_NOTIFICATION messages are published by OLTP systems
      * after it commits a transaction. (Batch applications can also
      * publisher TYPE_NOTIFICATION messages.)
      * <p>TYPE_REQUEST messages are <i>point-to-point</i> routed client
      * requests for service.  (<i>Point-to-point</i> messages are sent to a
      * single recipient, and have the recipient identified in the message
      * header, verus <i>publish-subscribe</i> messages that have no recipeint and
      * are published to all subscribers to the message topic.)
      * @see #setRecipient(String t)
      *
      */
    void setType(char t);


   /**
     * Sets the time the message was publisherd.  This
     * function is called by the message publisher after
     * client submits the message for delivery.
     * <p><B>NOTE</B> This value should be set by the underlying
     * client middleware, and not the application.
     *
     * @param n the time as a long.
     */
   void setTimeStamp(long n);


   /**
     * Sets the sequence number for this message.
     * <p><B>NOTE</B> This value should be set by the underlying
     * client middleware, and not the application.
     *
     * @param value the new sequence number
     */
  void setSequenceNumber(long n);

  /**
    * Returns a timestamp indicating when the message was sent.
    *
    * @return  A long containing the time in milli-seconds.
    */
  long getTimeStamp();


  /**
    * Returns the relative sequence number of this message.
    * When a message is sent by a client, it denotes the relative
    * sending sequence.  When a message is received, it denotes
    * the relative sequence of all messages fetched by the recipient,
    * regardless of the sender.
    *
    * @return  A long identifying the message sequence.
    */
    long getSequenceNumber();

    /**
      *  Identifies the body format as either dataset, text or binary.
      *  Valid values include
      * <p>  MessageAttributes.FORMAT_UNKNOWN
      * <p>  MessageAttributes.FORMAT_TEXT
      * <p>  MessageAttributes.FORMAT_DATASET
      * <p>  MessageAttributes.FORMAT_BINARY
      * <p> ERMessages default to a format of
      * MessageAttributes.FORMAT_DATASET.
      */
    char getFormat();

    /**
      *  Sets the body format identifier.
      *  Body formats are either dataset, text or binary.
      *  Valid values include
      * <p>  MessageAttributes.FORMAT_UNKNOWN
      * <p>  MessageAttributes.FORMAT_TEXT
      * <p>  MessageAttributes.FORMAT_DATASET
      * <p>  MessageAttributes.FORMAT_BINARY
      *
      * @exception IllegalArgumentException if attempt to set value
      *            other than MessageAttributes.FORMAT_DATASET
      */
      
      void setFormat(char t);

   // --------------------- TEST IF ATTRIBUTE IS SET ----------------------

   /**
     * Determine if there is a recipient identifier value.
     *
     *
     * @return  true if there is a non-null recipient identifier.
     */
   boolean hasAssignedRecipient();


   /**
     * Determines if there is a topic identifier value.
     *
     * @return  true if there is a non-null topic identifier.
     */
   boolean hasAssignedTopic();


   /**
     * Determines if there is a correlation identifier value.
     *
     * @return  true if there is a non-null numeric correlation identifier.
     */
   boolean hasAssignedCorrelationNumber();


   /**
     * Determines if there is a key value identifier.
     *
     * @return  true if there is a non-null key identifier.
     */
   boolean hasAssignedKey();


  /**
    * Determines if there is an originator identifier.
    *
    * @return  true if the originator has been set to a non-null value, else false
    */

   boolean hasAssignedOriginator();


   String toString();


  /**
    * Returns the number of positions in the message body (including those that are empty)
    *
    * @return  the number of body elements
    * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
    */

   int getElementCount();

   void setElementCount(int sz);

   Object clone();
   }














