/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */
 

package osmq.messages;

import java.util.*;
import java.math.*;
import java.io.Serializable;

import osmq.records.*;
import osmq.util.Immutable;
import osmq.util.ImmutableException;

/**
  * An entity relational message (ERMessage) contains entity data
  * values, and supports mapping those data values to/from Java
  * data types, and routing them as a unit to remote processes.  
  */
public class ERMessageBean implements ERMessage, Cloneable, Serializable
   {
   private static MessageStringSplitter _splitter;
   MessageHeader header;
   DataRecord body;
   
   // this is the main constructor -- the one all the others must call
   public ERMessageBean(MessageHeader header, DataRecord body)
      {
      this.body = body;
      this.header = header;
      
      // Identifies the the format as a ERMessage
      header.setFormat(MessageFormat.DATASET);
      }

    // this is the main constructor -- the one all the others must call
   public ERMessageBean(MessageHeader header, byte[] body)
      {
      this.body = new DataRecordBean(body);
      this.header = header;
      
      // Identifies the the format as a ERMessage
      header.setFormat(MessageFormat.DATASET);
      }

   public ERMessageBean()
      {
      this(new MessageHeader(), new DataRecordBean());
      }

   public ERMessageBean(int sz)
      {
      this(new MessageHeader(),
           new DataRecordBean(sz));
      }

   // ----------------------------------------------------
   // The remainder of the constructors are not public
   // ----------------------------------------------------
   ERMessageBean(String headerstring, String bodystring)
      {
      this(new MessageHeader(headerstring),
           new DataRecordBean(bodystring));
      }
         
   ERMessageBean(String[] value)
      {
      this(value[0], value[1]);
      }

   ERMessageBean(String s)
      {
      // Splits the string into two strings, and passes
      // a string array to the previous constructor
      this(getSplitter().split(s, new String[2]));
      }
	 
   ERMessageBean(MessageHeader h, String s)
      {
      this(h, new DataRecordBean(s));
      }

   ERMessageBean(MessageHeader header, int sz)
      {
      this(header, new DataRecordBean(sz));
      }

   /**
     * Sets the base index used to reference elements in the message body.
     * If not set explicitly, this index defaults to 1.
     * @param value the new base index value
  	 */
  public void setOffsetIndex(int value)
      {
      body.setOffsetIndex(value);
      }

   /**
     * Returns the base index currently used to reference elements in the message body.
     * @see #setOffsetIndex(int)
	   */
   public int getOffsetIndex()
      {
      return body.getOffsetIndex();
      }

   /**
     * Returns true if this is an immutable message.
     * @see #setImmutable()
     */
   public boolean isImmutable()
      {
      return body.isImmutable();
      }


   /**
     * Clears the value at the designated position, setting it to null (or zero if numeric)
     * @param posn the first position is 1, the second is 2, ...
     * @exception ImmutableDataException if the message is immutable.
     * @see #clearBody()  
     */
   public void clear(int posn) 
      {
      body.clear(posn);
      }


   /**
     * Returns the value of the designated position in the message body.
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the value as an Object or null  if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a 
	   *             valid position.
     */
   public Object getObject(int posn)
	  {
	  return body.getObject(posn);
	  }
      
   /**
     * Returns the value at the designated position in the message body as a char
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the value as a char or 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */
   public char getChar(int posn)
	  {
	  return body.getChar(posn);
	  }


   /**
     * Returns the value at the designated position in the message body as a byte
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the value as a byte or 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */
   public byte getByte(int posn)
	  {
	  return body.getByte(posn);
	  }

   /**
     *  Returns the value at the designated position in the message body as a boolean
	 *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the value as a boolean or a null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */
   public boolean getBoolean(int posn)
	  {
	  return body.getBoolean(posn);
	  }

   /**
     *  Returns the value at the designated position in the message body as a float
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the position value as a double or 0.0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a 
	 *             valid position.
     */
   public float getFloat(int posn)
	  {
      return body.getFloat(posn);
	  }

   /**
     * Returns the value at the designated position in the message body as a double
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the position value as a double or 0.0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a 
	   *             valid position.
     */
  public double getDouble(int posn)
	 {
	 return body.getDouble(posn);
	 }

  /**
     *  Returns the value at the designated position in the message body as a BigDecimal
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the position value as a BigDecimal or null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     */
  public BigDecimal getBigDecimal(int posn)
	 {
	 return body.getBigDecimal(posn);
	 }


  // -------------------- POSITIONAL GETTERS ----------------------

   /**
     * Returns the value at the designated position in the message body as an integer.
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the value as an int or a 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */
  public int getInt(int posn)
	 {
	 return body.getInt(posn);
	 }

   /**
     *  Returns the value at the designated position in the message body as a long integer
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the position value as a short or 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if columnIndex is not a valid element position.
     */
  public long getLong(int posn)
	 {
	 return body.getLong(posn);
	 }

   /**
     *  Returns the value at the designated position in the message body as a short integer
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the position value as a short or 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if columnIndex is not a valid element position.
     */
  public short getShort(int posn)
	  {
    return body.getShort(posn);
    }

   /**
     * Returns the value of the designated position in the message body as a String
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the position value as a String or a null String if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */
  public String getString(int posn)
	  {
      return body.getString(posn);
	  }

  /**
    *  Returns the value at the designated position in the message body as a String 
	*  with all characters forced into upper case.
    *
    * @param posn the first position is 1, the second is 2, ...
    * @return  the value as a String or a null if the position is empty.
    * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid element position.
    */
  public String getUpperCaseString(int posn)
     {
     return body.getUpperCaseString(posn);
     }

   /**
     *  Returns the value at the designated position in the message body as a java.util.Date
     *
     * @param posn the first position is 1, the second is 2, ...
     * @return  the position value as a java.util.Date, or null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if columnIndex is not a valid element position.
     * @see #setDate(int, Date)
     */
  public Date getDate(int posn)
     {
     return body.getDate(posn);
     }

   /**
     *  Returns the value at the designated position in the message object as an SQL date
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the position value as a java.sql.Date object, or null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
	 *             valid position.
     * @see #setDateString(int, String)
     */
  public java.sql.Date getSQLDate(int posn)
     {
     return body.getSQLDate(posn);
     }

  /**
     * Returns the value at the designated position in the message object
     * as a formatted date string in CCYY-MM-DD format.
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the position value as a String or null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
	 *             valid position.
     * @see #setDateString(int, String)
     */
  public String getFormattedDateString(int posn)
     {return body.getFormattedDateString(posn);}


  /**
     * Returns the value at the designated position in the message object
     * as an unformatted date string in CCYYMMDD format.  (This is the normal
     * storage format for a date).
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the position value as a String or null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
	 *             valid position.
     * @see #setDateString(int, String)
     */
  public String getUnformattedDateString(int posn)
     {
     return body.getUnformattedDateString(posn);
     }

  /**
    * Determines whether a position has a value (versus null)
    *
    * @param posn the position containing the value to be tested
    * @return  <tt>true</tt> of there is a non-null value at position p, else <i>false</i>
    * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
    */
  public boolean hasAValue(int posn)
	   {
	   return body.hasAValue(posn);
	   }

  /**
    * Determines if a position has no value (is empty).  Alias for isNull(int)
    *
    * @param posn the position containing the value to be tested
    * @return  <tt>true</tt> if there is a null value (no value) at position p, else <tt>false</tt>
    * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
    */
  public boolean isEmpty(int posn)
     {
     return body.isEmpty(posn);
     }

  /**
    * Determines if a position has no value (has a null value).  Alias for isEmpty(int)
    *
    * @param posn the position containing the value to be tested
    * @return  <tt>true</tt> if there is a null value (no value) at position p, else <tt>false</tt>
    * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
    */
  public boolean isNull(int posn)
     {
     return body.isNull(posn);
     }

   /**
     * Sets the value at the designated position in the message body
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if p is not a valid element position.
     * @exception ImmutableException if the message object is immutable.
     */
  public void setObject(int posn, Object value)
	   {
     body.setObject(posn, value);
	   }

   /**
     * Sets the element at the designated position in the message body to a String value
     *
     * @param posn the element's relative position (1 to n)
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if p is not a valid element position.
     * @exception ImmutableException if the message object is immutable.
     */
  public void setString(int posn, String s) 
	   {
	   body.setString(posn, s);
	   }

  public void setTrimmedString(int posn, String s)
     {
     body.setTrimmedString(posn, s);
     }

   /**
     * Sets the value at the designated position in the message body to null
     *
     * @param posn the first position is 1, the second is 2, ...
     * @exception  ArrayIndexOutOfBoundsException if p is not a valid element position.
     * @exception ImmutableException if the message object is immutable.
     */
  public void setNull(int posn)
     {
     body.setNull(posn);
     }

   /**
     * Sets the element at the designated position in the message body to a byte value
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the message is immutable.
     */
  public void setByte(int posn, byte value)
	  {
	  body.setByte(posn, value);
	  }

  
   /**
     * Sets the element at the designated position in the message body to a boolean value
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the message is immutable.
     */
  public void setBoolean(int posn, boolean value)
	  {
	  body.setBoolean(posn, value);
	  }

   /**
     * Sets the element at the designated position in the message body to a char value
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the message is immutable.
     */
  public void setChar(int posn, char value)  
	  {
	  body.setChar(posn, value);
	  }

   /**
     * Sets the element at the designated position in the message body to a float value
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid  position.
	   * @exception ImmutableException if the message is immutable.
     */
   public void setFloat(int posn, float value)
	   {
	   body.setFloat(posn, value);
	   }

   /**
     * Sets the element at the designated position in the message body to a double value
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid  position.
     * @exception ImmutableException if the message is immutable.
     */
  public void setDouble(int posn, double value)
	  {
	  body.setDouble(posn, value);
	  }


  public void setDecimalString(int posn, String value)
     {
     body.setDecimalString(posn, value);
     }

   /**
     * Sets the element at the designated position in the message body to a BigDecimal value
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid  position.
     * @exception ImmutableException if the message is immutable.
     */
  public void setBigDecimal(int posn, BigDecimal value)
     {
     body.setBigDecimal(posn, value);
     }

    /**
     * Sets the element at the designated position in the message body to a Date value
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid  position.
	   * @exception ImmutableException if the message is immutable.
     */
  public void setDate(int posn, Date value)
    {
    body.setDate(posn, value);
    }

   /**
     * Sets the element at the designated position in the message body to an int value
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
	   * @exception ImmutableException if the message is immutable.
     */
  public void setInt(int posn, int value)
	  {
	  body.setInt(posn, value);
	  }

   /**
     * Sets the element at the designated position in the message body to a long value
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the message is immutable.
     */
  public void setLong(int posn, long value)
	  {
	  body.setLong(posn, value);
	  }

   /**
     * Sets the element at the designated position in the message body to a short value
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the message is immutable.
     */
  public void setShort(int posn, short value)
	 {
	 body.setShort(posn, value);
	 }

   /**
     * Sets the element at the designated position in the message body to an
	   * upper-case String value
     *
     * @param posn the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the message object is immutable.
     */
  public void setUpperCaseString(int posn, String value)
	 {
  	 body.setUpperCaseString(posn, value);
  	 }

  /**
     * Sets the element at the designated position in the message body to an date value.
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value in CCYYMMDD or CCYY-MM-DD format.  (If the string
     *        contains hyphens, they are removed.)
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
	 * @exception ImmutableException if the message object is immutable.
     */
  public void setDateString(int posn, String value)
     {
     body.setDateString(posn, value);
     }

   /**
     * Clear all the body elements in the ERMessage object to null (or zero if numeric)
     * @exception ImmutableException if the message is immutable.
     */
  public void clearBody()
     {
     body.clearAll();
     }

   /**
     * Clear all the header elements in the ERMessage object to null (or zero if numeric)
     * @exception ImmutableException if the message is immutable.
     */
  public void clearHeader()
     {
     header.clearAll();
     }

  private void createAHeader()
     {
     header = new MessageHeader();
     }

  /**
    * Sets the message state to "immutable."  The values of
    * an immutable message cannot be modified.  Any attempt to
    * modify those values will result in an ImmutableException being
    * thrown.
    * @see osmq.util.ImmutableException
    */
  public void setImmutable()
     {
     body.setImmutable();
     if(header != null)
       header.setImmutable();
     }
     
   /**
     * Sets the message header values.
     * @exception ImmutableDataException if the message is immutable.
     */
   public void setHeader(MessageHeader header)
     {
     errorIfImmuatable();
     this.header = header;
     }

   /**
     * Returns the message header values.
     */
   public MessageHeader getHeader()
     {
     return header;
     }

   /**
     * Returns the header data as an array of bytes.
     */
   public byte[] getHeaderBytes()
      {
      return header.getBytes();
      }

   /**
     * Sets the header data from an array of bytes.
     * @exception ImmutableDataException if the message is immutable.
     */
   public void setHeaderBytes(byte[] value) 
      {
      errorIfImmuatable();
      header = new MessageHeader(value);
      }

   /**
     * Returns the client (body) data as an array of bytes.
     */
   public byte[] getClientBytes()
      {
      return body.getBytes();
      }
      
   /**
     * Sets the client (body) data from values in a message body.
     * Changes made to the DataRecord will change the corresponding
     * values in the message body.
     * @exception ImmutableDataException if the message is immutable.
     */
   public void setClientBytes(DataRecord ds)
      {
      errorIfImmuatable();
      this.body = ds;
      }

   /**
     * Sets the client (body) data from an array of bytes.
     * @exception ImmutableDataException if the message is immutable.
     */
   public void setClientBytes(byte[] data)
      {
      errorIfImmuatable();
      body.setBytes(data);
      }

   /**
     * Returns the message header and body as an array of bytes.
     */
   public byte[] getMessageAsBytes()
      {
      return(header.getText() +
             getElementsAsString()).getBytes();
      }

   /**
     * Returns the message body as a delimited String.
     */
   public String getElementsAsString()
      {
      return body.toString();
      }
      
   /** Returns <i>true</i> if the <i>lastfetched</i> element was a null object, or
     * an object with a string length of zero.  (The <i>last fetched</i>
     * is the element referenced by the last <code>getXXX()</code> method call.)
     */
  public boolean wasNull()
     {
     return body.wasNull();
     }

   /**
     * Returns the topic of the message.
     *
     * @return  A string containing the topic. The topic defines the
     * message contents for a subscriber (number of elements and the
     * data type of each element).
     */
   public String getTopic()
      {
      return header.getTopic();
      }

   /**
     * Sets the topic that identifies the message contents.
     * @exception ImmutableDataException if the message is immutable.
     */
   public void setTopic(String topic)
      {
      header.setTopic(topic);
      }

   /**
     * Returns the unique identifier for the data contained in the message.
     * (This "key" is relative to the the topic.)
     *
     * @return  A string containing the unique identifier.
     */
   public String getKey()
      {
      return header.getKey();
      }

   /**
     * Sets the unique identifier for the data contained in the message.
     * (This "key" is relative to the the topic.)
     *
     * @param s the unique key value
     */
   public void setKey(String s)
      {
      header.setKey(s);
      }

   /**
     * Sets the unique identifier for the data contained in the message.
     * (This "key" is relative to the the topic.)
     *
     * @param n the unique key value
     */
   public void setKeyAsInt(int n)
      {
      header.setKeyAsInt(n);
      }

   public int getKeyAsInt()
      {
      return header.getKeyAsInt();
      }

   /**
     * Returns the identifier of the event or activity that caused the
     * transaction to be generated.  This is a value that is based on the topic.
     * <p>
     * Standard events used for down-stream subscribers that maintain
     * databases are ERMessage.TRANSACTION_ADD, ERMessage.TRANSACTION_DELETE
     * and ERMessage.and TRANSACTION_UPDATE.
     *
     * @return  A string containing the transaction event.
     */
   public String getTransaction()
      {
      return header.getTransaction();
      }

   /**
    * Returns the first character of the event identifer, or a null character if
    * there is no event identifer
    */
    public char getTransactionAsChar()
       {
       return header.getTransactionAsChar();
       }

  /**
    * Sets the transaction identifer.  This identifier is useful to
    * subscribers that perform database updates based on topical messages.
    * (The identifiers typically translate into add/update/delete functions)
    */
    public void setTransaction(String e)
       {
       header.setTransaction(e);
       }

  /**
    * Sets the transaction identifer to a char value.
    * @see #setTransaction(String)
    */
    public void setTransactionAsChar(char c)
       {
       header.setTransactionAsChar(c);
       }


    /**
      * Gets the identifier used to identify a group of transactions
      * as related.  The relationship is likely a unit-of-work that
      * resulted in tranaction-related topics being generated.
      * This is an optional header attribute.
      *
      * @return  An int containing the correlation identifier.
      */
    public long getCorrelationNumber()
       {
       return header.getCorrelationNumber();
       }

    /**
      * Sets the identifier used to identify a group of transactions
      * as related.  The relationship is likely a unit-of-work that
      * resulted in tranaction-related topics being generated.
      * This is an optional header attribute, and is set automatically
      * by the Publisher.
      *
      * @param n the correlation identifier
      */
    public void setCorrelationNumber(long n)
       {
       header.setCorrelationNumber(n);
       }

    public String getCorrelationString()
       {
       return header.getCorrelationString();
       }

    /**
      * Sets the identifier used to identify a group of transactions
      * as related.
      * @see #setCorrelationNumber(long)
      *
      * @param s the correlation identifier
      */
    public void setCorrelationString(String s)
       {
       header.setCorrelationString(s);
       }

    /**
      * Returns the identifier of the message sender (publisher).  The
      * sender identifies a unique client process (instance).
      *
      * @return  A string containing the message sender.
      */
    public String getOriginator()
       {
       return  header.getOriginator();
       }

    /**
      * Sets the identifier of the message sender (publisher).  The
      * sender identifies a unique client process (instance).
      * <p> Normally, this value is not set by an application, and
      * defaults to the client identifier used to attach to the message broker.
      *
      * @param t the message originator (sender)
      */
    public void setOriginator(String t)
       {
       header.setOriginator(t);
       }

    /**
      * Returns the identifier of the message recipient, or
      * null if the message is intended for all subscriber to the
      * message topic.
      *
      * @return  A string containing the recipient identifier.
      */
    public String getRecipient()
       {
       return header.getRecipient();
       }

    /**
      * Sets the unique recipient that will receive the message.
      * If a recipient value is set, only that recipient will receive
      * the message. (Point-to-point messaging).  Setting the recipient
      * should be reserved for clients requesting a service (type of
      * TYPE_REQUEST) and for OLTP systems that receive a request that
      * is being rejected.  In the latter case, the OLTP system would send
      * a point-to-point rejection message to the originator of the request.
      *
      * @param value the intended message recipient
      */
    public void setRecipient(String t)
      {
      header.setRecipient(t);
      }

    /**
      * Identifies whether the event type is a MessageAttributes.NOTIFICATION
      * or MessageAttributes.REQUEST.
      *
      * @return  A char containing either the value TYPE_NOTIFICATION or
      * TYPE_REQUEST.
      */
    public char getType()
      {
      return header.getType();
      }

    /**
      * Sets the message type. The message type is the category
      * of event that caused a message to be generated.
      * Standard typess include  MessageAttributes.TYPE_NOTIFICATION
      * and MessageAttributes.TYPE_REQUEST.
      * <p>TYPE_NOTIFICATION messages are published by OLTP systems
      * after it commits a transaction. (Batch applications can also
      * publisher TYPE_NOTIFICATION messages.)
      * <p>TYPE_REQUEST messages are <i>point-to-point</i> routed client
      * requests for service.  (<i>Point-to-point</i> messages are sent to a
      * single recipient, and have the recipient identified in the message
      * header, verus <i>publish-subscribe</i> messages that have no recipeint and
      * are published to all subscribers to the message topic.)
      * @see #setRecipient(String t)
      *
      */
    public void setType(char t)
       {
       header.setType(t);
       }

   /**
     * Sets the time the message was publisherd.  This
     * function is called by the message publisher after
     * client submits the message for delivery.
     * <p><B>NOTE</B> This value should be set by the underlying
     * client middleware, and not the application.
     *
     * @param n the time as a long.
     */
   public void setTimeStamp(long n)
      {
      header.setTimeStamp(n);
      }

   /**
     * Sets the sequence number for this message.
     * <p><B>NOTE</B> This value should be set by the underlying
     * client middleware, and not the application.
     *
     * @param value the new sequence number
     */
  public void setSequenceNumber(long n)
      {
      header.setSequenceNumber(n);
      }

  /**
    * Returns a timestamp indicating when the message was sent.
    *
    * @return  A long containing the time in milli-seconds.
    */
  public long getTimeStamp()
     {
     return header.getTimeStamp();
     }

  /**
    * Returns the relative sequence number of this message.
    * When a message is sent by a client, it denotes the relative
    * sending sequence.  When a message is received, it denotes
    * the relative sequence of all messages fetched by the recipient,
    * regardless of the sender.
    *
    * @return  A long identifying the message sequence.
    */
   public  long getSequenceNumber()
      {
      return header.getSequenceNumber();
      }

    /**
      *  Identifies the body format as either dataset, text or binary.
      *  Valid values include
      * <p>  MessageAttributes.FORMAT_UNKNOWN
      * <p>  MessageAttributes.FORMAT_TEXT
      * <p>  MessageAttributes.FORMAT_DATASET
      * <p>  MessageAttributes.FORMAT_BINARY
      * <p> ERMessages default to a format of
      * MessageAttributes.FORMAT_DATASET.
      */
    public char getFormat()
       {
       return header.getFormat();
       }

    /**
      *  Sets the body format identifier.
      *  Body formats are either dataset, text or binary.
      *  Valid values include
      * <p>  MessageAttributes.FORMAT_UNKNOWN
      * <p>  MessageAttributes.FORMAT_TEXT
      * <p>  MessageAttributes.FORMAT_DATASET
      * <p>  MessageAttributes.FORMAT_BINARY
      *
      * @exception IllegalArgumentException if attempt to set value
      *            other than MessageAttributes.FORMAT_DATASET
      */
      
      public void setFormat(char t)
         {
         if(t != MessageFormat.DATASET)
            throw new IllegalArgumentException("Attempt to set incorrect format");
         header.setFormat(t);
         }


   // --------------------- TEST IF ATTRIBUTE IS SET ----------------------

   /**
     * Determine if there is a recipient identifier value.
     *
     *
     * @return  true if there is a non-null recipient identifier.
     */
  public boolean hasAssignedRecipient()
      {
      return header.hasAssignedRecipient();
      }

   /**
     * Determines if there is a topic identifier value.
     *
     * @return  true if there is a non-null topic identifier.
     */
   public boolean hasAssignedTopic()
      {
      return header.hasAssignedTopic();
      }

   /**
     * Determines if there is a correlation identifier value.
     *
     * @return  true if there is a non-null numeric correlation identifier.
     */
   public boolean hasAssignedCorrelationNumber()
      {
      return header.hasAssignedCorrelationNumber();
      }

   /**
     * Determines if there is a key value identifier.
     *
     * @return  true if there is a non-null key identifier.
     */
   public boolean hasAssignedKey()
      {
      return header.hasAssignedKey();
      }

    /**
    * Determines whether a position has a value (versus null)
    *
    * @param p the position containing the value to be tested
    * @return  true of there is a non-null value at position p, else false
    * @exception  ArrayIndexOutOfBoundsException if p is not a valid element position.
    */

   public boolean hasAssignedOriginator()
      {
      return header.hasAssignedOriginator();
      }

   public String toString()
      {
      return header.getText() + body.getText();
      }


  /**
    * Returns the number of positions in the DataRecord (including those that are empty)
    *
    * @return  the number of DataRecord elements
    * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
    */

  public int getElementCount()
	   {
     return body.getElementCount();
     }

  public void setElementCount(int sz)
     {
     errorIfImmuatable();
     body = new  DataRecordBean(sz);
     }

   public Object clone()
	    {
        ERMessageBean dcm;
        try{dcm = (ERMessageBean) super.clone();}
        catch(CloneNotSupportedException e){return this;}
        dcm.header = (MessageHeader) header.clone();
        dcm.body = (DataRecord) body.clone();
        return dcm;
        }

   private static MessageStringSplitter getSplitter()
      {
      if(_splitter == null)
         {
         synchronized(ERMessageBean.class)
            {
            if(_splitter == null)
               _splitter = new MessageStringSplitter();
            }
         }
      return _splitter;
      }

   private void errorIfImmuatable()
      {
       if(isImmutable())
         throw new ImmutableException("Message is immutable");
      }
   }














