/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */


package osmq.messages;

import java.util.*;
import java.io.Serializable;

import osmq.records.*;
import osmq.util.ImmutableImpl;


public class MessageHeader extends ImmutableImpl implements MessageAttributes,
                                                 Cloneable, Serializable
   {
   private DataRecord ds;
   
   public MessageHeader()
      {
      this(new DataRecordBean(MessageHeader.TOTAL_ATTRIBUTES));
      }
        
   public MessageHeader(DataRecord ds)
      {this.ds = ds;}

   public MessageHeader(String s)
      {
      this(new DataRecordBean(s));
      }

   public MessageHeader(byte[] value)
      {
      this(new DataRecordBean(value));
      }

    /**
      * Gets the identifier used to identify a group of transactions
      * as related.  The relationship is likely a unit-of-work that
      * resulted in tranaction-related topics being generated.
      * This is an optional header attribute.
      *
      * @return  An int containing the correlation identifier.
      */
   public long getCorrelationNumber()
      {
      return ds.getLong(CORRELATION_ID);
      }

   public void setCorrelationNumber(long n)
      {
      ds.setLong(CORRELATION_ID, n);
      }

   public String getCorrelationString()
      {
      return ds.getString(CORRELATION_ID);
      }

   public void setCorrelationString(String s)
      {
      ds.setString(CORRELATION_ID, s);
      }


  /////////////////////////////////////////////////////////////////
  // TOPIC
  /////////////////////////////////////////////////////////////////
   /**
     * Returns the topic of the message.
     *
     * @return  A string containing the topic. The topic defines the
     * message contents for a subscriber (number of elements and the
     * data type of each element).
     */
   public String getTopic()
      {
      return ds.getString(TOPIC);
      }

   public void setTopic(String t)
      {
      ds.setString(TOPIC, t);
      }

   /**
     * Returns the unique identifier for the data contained in the message.
     * (This "key" is relative to the the topic.)
     *
     * @return  A string containing the unique identifier.
     */
   public String getKey()
      {
      return ds.getString(DATA_KEY);
      }

   public void setKey(String t)
      {
      ds.setUpperCaseString(DATA_KEY, t);
      }

   public void setKeyAsInt(int n)
      {
      ds.setInt(DATA_KEY, n);
      }

   public int getKeyAsInt()
      {
      return ds.getInt(DATA_KEY);
      }

   /**
     * Returns the identifier of the event or activity that caused the
     * transaction to be generated.  This is a value that is based on the topic.
     * <p>
     * Standard events used for down-stream subscribers that maintain
     * databases are ERMessage.EVENT_ADD, ERMessage.EVENT_DELETE
     * and ERMessage.and EVENT_UPDATE.
     *
     * @return  A string containing the transaction event.
     */
   public String getTransaction()
      {
	    return ds.getString(TRANSACTION);
	    }

  /**
    * Returns the first character of the event identifer, or a null character if
    * there is no event identifer
    */
   public char getTransactionAsChar()
      {
	    return ds.getChar(TRANSACTION);
	    }

   public void setTransaction(String t)
      {
	    ds.setString(TRANSACTION, t);
	    }
    
   public void setTransactionAsChar(char c)
      {
	    ds.setChar(TRANSACTION, c);
	    }

    /**
      * Returns the identifier of the message sender (publisher).  The
      * sender identifies a unique client process (instance).
      *
      * @return  A string containing the message sender.
      */
   public String getOriginator()
      {
	    return ds.getString(ORIGINATOR);
	    }


    /**
      * Sets the identifier of the message sender/originator.
      * Setting the originator should be performed when a message is published,
      * and this value defaults to the client identifier used to attach to the
      * message server.
      * <p><b>NOTE</b> This method should not be called by an application, since
      * the value will be overridden by the message publisher adapter.
      *
      * @param value the sender/originator of the message
      */
   public void setOriginator(String t)
      {
      ds.setString(ORIGINATOR, t);
      }

    /**
      * Identifies whether the event type is a NOTIFICATION or REQUEST .
      *
      * @return  A char containing either the value TYPE_NOTIFICATION or
      *          TYPE_REQUEST.
      */
   public char getType()
     {
	   return ds.getChar(TYPE);
 	   }

    /**
      * Sets the message type. The message type is the category
      * of event that caused a message to be generated.
      * Standard types include  TYPE_NOTIFICATION and TYPE_REQUEST.
      * <p>TYPE_NOTIFICATION messages are published by OLTP systems
      * after it commits a transaction. (Batch applications can also
      * publisher TYPE_NOTIFICATION messages.)
      * <p>TYPE_REQUEST messages are <i>point-to-point</i> routed client
      * requests for service.  (<i>Point-to-point</i> messages are sent to a
      * single recipient, and have the recipient identified in the message
      * header, verus <i>publish-subscribe</i> messages that have no recipeint and
      * are published to all subscribers to the message topic.)
      * @see #setRecipient(String t)
      *
      */
   public void setType(char t)
     {
 	   ds.setChar(TYPE, t);
 	   }

    /**
      * Returns the identifier of the message recipient, or
      * null if the message is intended for all subscriber to the
      * message topic.
      *
      * @return  A string containing the recipient identifier.
      */
   public String getRecipient()
     {
	   return ds.getString(RECIPIENT);
 	   }

    /**
      * Sets the unique recipient that will receive the message.
      * If a recipient value is set, only that recipient will receive
      * the message. (Point-to-point messaging).  Setting the recipient
      * should be reserved for clients requesting a service (type of
      * TYPE_REQUEST) and for OLTP systems that receive a request that
      * is being rejected.  In the latter case, the OLTP system would send
      * a point-to-point rejection message to the originator of the request.
      *
      * @param value the intended message recipient
      */
   public void setRecipient(String s)
       {
 	   ds.setString(RECIPIENT, s);
 	   }


  ////////////////////////////////////////////////////////////
  // MESSAGE CONTENT TYPE
  ////////////////////////////////////////////////////////////

   public char getFormat()
      {return ds.getChar(CONTENT_FORMAT);}

   public void setFormat(char t)
      {ds.setChar(CONTENT_FORMAT, t);}


  ////////////////////////////////////////////////////////////
  // MESSAGE TIME STAMP
  ////////////////////////////////////////////////////////////

  /**
    * Returns a timestamp indicating when the message was sent.
    *
    * @return  A long containing the time in milli-seconds.
    */
   public long getTimeStamp()
      {return ds.getLong(SEND_TIMESTAMP);}

   /**
     * Sets the time the message was publisherd.  This
     * function is called by the message publisher after
     * client submits the message for delivery.
     * <p><B>NOTE</B> This value should be set by the underlying
     * client middleware, and not the application.
     *
     * @param n the time as a long.
     */
   public void setTimeStamp(long n)
      {ds.setLong(SEND_TIMESTAMP, n);}
     
  /**
    * Returns the relative sequence number of this message.
    * When a message is sent by a client, it denotes the relative
    * sending sequence.  When a message is received, it denotes
    * the relative sequence of all messages fetched by the recipient,
    * regardless of the sender.
    *
    * @return  A long identifying the message sequence.
    */
   public long getSequenceNumber()
      {return ds.getLong(SENDER_SEQ);}

   /**
     * Sets the sequence number for this message.
     * <p><B>NOTE</B> This value should be set by the underlying
     * client middleware, and not the application.
     *
     * @param value the new sequence number
     */
   public void setSequenceNumber(long n)
      {ds.setLong(SENDER_SEQ, n);}

      
   /**
     * Determine if there is a recipient identifier value.
     *
     * @return  true if there is a non-null recipient identifier.
     */
    public boolean hasAssignedRecipient()
       {
	     return ds.hasAValue(RECIPIENT);
	     }

   /**
     * Determine if there is a topic identifier value.
     *
     * @return  true if there is a non-null topic identifier.
     */
    public boolean hasAssignedTopic()
       {
   	   return ds.hasAValue(TOPIC);
	     }

    public boolean hasAssignedKey()
       {
	     return  ds.hasAValue(DATA_KEY);
	     }
	 
    public boolean hasAssignedCorrelationNumber()
       {
	     return ds.hasAValue(CORRELATION_ID);
	     }

    public boolean hasAssignedOriginator()
       {
       return ds.hasAValue(ORIGINATOR);
       }

    public byte[] getBytes()
       {return ds.getBytes();}

    public String getText()
       {return ds.getText();}

    public void clearAll()
       {ds.clearAll();}

   public Object clone()
     {
     if(this.isImmutable())
         return this;
     MessageHeader dcm = (MessageHeader) super.clone();
     //catch(CloneNotSupportedException e){return this;}
     dcm.ds = (DataRecord) ds.clone();
     return dcm;

     //return (MessageHeader) super.clone();
     //try{return (DefaultMessageAttributes) super.clone();}
     //catch(CloneNotSupportedException e){return this;}
     }
   }














