/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */
 
package osmq.mque;

import java.util.*;
import java.io.*;

import osmq.io.PacketFileInputStream;
import osmq.io.PacketFileOutputStream;

/**
  * A Queue implementation that utilizes a persistent data file
  */
class DiskFifo extends ReadableFifo
  {
  private byte[] rv;

  // output file was previously opened
  private boolean didopen = false;
  private long  pulled = 0;
  private static int fifoID = 0;

  private boolean deleteOnClose = true;
  transient PacketFileInputStream inf = null;
  transient PacketFileOutputStream outf = null;

  private boolean   infIsOpen = false;
  private boolean   outfIsOpen = false;
  private String    filename = null;

  DiskFifo()
     {
     // unique fifo instance name                         
     filename = new String("df" + fifoID++); 
     }

  private void openInf() throws IOException
     {
     if(inf != null)
        inf.close();
     inf = new PacketFileInputStream(filename);
     infIsOpen = true;
     }

  private void openOutf() throws IOException
     {
     if(outf != null)
        outf.close();
     // The first open destroys any prior file,
     // subsequent ones are append
     outf = new PacketFileOutputStream(filename, didopen);
     didopen = true;
     outfIsOpen = true;
     }

  private void closeInf()
     {
      try{
         if(infIsOpen)
            inf.close();
         }
      catch(IOException e){}
      finally{inf = null;}
     }

  private void closeOutf()
     {
     try{if(outfIsOpen)
        outf.close();}
     catch(IOException ignore){}
     finally{outf = null;}
     }

  /**
    * Sets the Queue's state to <i>closed</i> so that no objects can be
    * written to the fifo.
    * @see #isClosed()
    */
   protected void closeImpl()
     {
     closeInf();
     closeOutf();
     zapFile();
     }

   private void zapFile()
     {
     File tmp = new  File(filename);
     tmp.delete();
     }

   protected void finalize() throws Throwable
      {
      close();
      }



  /**
    * Places an object at the end of the Queue.
    * @exception IOException if the Queue is in a <i>closed</i> state.
    */
  private void writeBytes(byte[] b) throws IOException
     {
     if(!outfIsOpen)
         openOutf();

      outf.writeBytes(b);

      if(!infIsOpen)
         openInf();
     incrementCount();
     }

   /**
    * Returns the next object (FIFO order) from the Queue, and removes that
	  * object from the Queue. If the Queue is empty, the function blocks until
    * it is notified that an object is available.
    * @return the next object from the Queue (FIFO order), or null if none is waiting.
    * @exception QueueException thrown if Queue is in a closed state AND the Queue is empty.
    * @exception QueueException thrown if Queue is in a <i>halt-reads</i> state.
    * @see #isClosed()
    * @see #isHaltReads()
    */
  public byte[] readBytes() throws IOException
     {
     if(size() < 1)
        throw new IOException("Fifo is empty");

     if(inf == null)
        openInf();
     rv = inf.readBytes();
     decrementCount();
     ++pulled;
     return rv;
     }

  public void appendFrom(ReadableFifo f) throws IOException
     {
     while(!f.isEmpty())
         writeBytes(f.readBytes());
     outf.flush();
     }

  public long getPulledCount()
     {return pulled;}


  public static void main(String[] args)
    {
    DiskFifo df = new DiskFifo();
    int i = 0;
    int cv = 0;
    try{
       //System.out.println("spill name = " + df.getSpillName());

       for(int ctr = 0; ctr < 100; ++ctr)
           {
           for(i = 0; i < 100; ++ i)
             df.writeBytes(new String("" + cv++).getBytes());

           for(i = 0; i < 99; ++ i)
              System.out.println(new String(df.readBytes()));
       //while(!df.isEmpty())

           }
      while(!df.isEmpty())
         System.out.println(new String(df.readBytes()));

       df.close();
       }
    catch(Exception e){}
    }
  }

