/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */

package osmq.net;

import java.io.*;
import java.net.*;
import java.util.*;
import osmq.io.*;
import osmq.util.*;

/**
  * Client stream for communicating with a remote service.
  */
public class SocketStreamFilter  extends ExceptionListenerImpl
   {
   final public static boolean IS_BINARY = true;

   private Socket  _socket = null;
   ReaderStream    _reader = null;
   WriterStream    _writer = null;
   private boolean _binarymode = false;
   private long    _written = 0;

   public SocketStreamFilter(Socket socket, boolean binarymode) throws IOException
       {
       setSocket(socket, binarymode);
       }

   public SocketStreamFilter(DiscoverySocket ds)  throws IOException
       {
       this(ds, ds.isBinaryMode());
       }

   public SocketStreamFilter (RemoteService es, ExceptionListener listener)
                                                 throws IOException
      {
      this(new DiscoverySocket(es));
      if(listener != null)
         setExceptionListener(listener);
      }
       
   /**
     * This method may be called to replace an exisiting socket.
     * This might be done by a client application if the remote
     * server fails, and there is a need to connect to an alternate
     * server.
     */
   public void setSocket(Socket socket, boolean binarymode) throws IOException
      {
      _socket = socket;
      _binarymode = binarymode;
      _reader = null;
      _writer = null;
      ensureOpen();
      }

   /**
     * This method may be called to replace an exisiting socket.
     * This might be done by a client application if the remote
     * server fails, and there is a need to connect to an alternate
     * server.
     */
    public void setSocket(DiscoverySocket socket) throws IOException
      {
      setSocket(socket, socket.isBinaryMode());
      }

   private ReaderStream getReaderStream() throws IOException
      {
      // reads from a BufferedInputStream
      if(_reader == null)
         {
         ensureOpen();
         _reader = createSocketReader(
            new BufferedInputStream(_socket.getInputStream()));
         }
      return _reader;
      }

   private WriterStream getWriterStream() throws IOException
      {
      if(_writer == null)
         {
         ensureOpen();
         _writer = createSocketWriter(
            _socket.getOutputStream());
         }
      return _writer;
      }

   public long getWrittenCount(){return _written;}

   public void printlnCounted(String s) throws IOException
        {
        println(s);
        _written++;
        }

   /**
     * Print a string to the socket stream
     * @param s the String to be printed
     * @exception  IOException if String size < 1 or there is a communications error
     */
   public synchronized void println(String s) throws IOException
       {
       try{
          if(s == null || s.length() < 1)
             throw new NetException("Attempt to print null String");
          ensureOpen();
          getWriterStream().println(s);
          }
       catch(IOException e)
           {
           this.onException(e);
           this.close();
           throw new NetException("SocketWrapper error in println "
                                 + e.getMessage());
           }
       }

     public void writeBytesCounted(byte[] v) throws IOException
        {
        writeBytes(v);
        _written++;
        }

     public synchronized void writeBytes(byte[] v) throws IOException
       {
       try{
          if(v == null || v.length < 1)
             throw new NetException("Attempt to print null String");
          ensureOpen();
          getWriterStream().writeBytes(v);
          }
       catch(IOException e)
           {
           this.onException(e);
           this.close();
           throw new NetException("SocketWrapper error in println "
                                 + e.getMessage());
           }
       }
   /**
     * Used to test active socket connections.  This sends
     * a value indicating a zero-sized string.
     */
   public synchronized void writeNull() throws IOException
       {
       try
          {
          ensureOpen();
          getWriterStream().writeNull();
          }
       catch(IOException e)
           {
           this.onException(e);
           this.close();
           throw new NetException("SocketWrapper error in writeNull "
                                 + e.getMessage());
           }
       }
       

   /**
     * Reads the next String from the socket input stream. This is
     * a blocking read -- it waits indefinitely for the next String.
     * <p> Note that the function ignores null length Strings.
     * @return  the value of the next byte packet as a String.
     * @exception IOException if there is a communications failure.
     */
   public synchronized String readln() throws IOException
       {
       try{
          ensureOpen();
          return getReaderStream().readln();
          }
       catch(IOException e)
           {
           this.onException(e);
           this.close();
           throw new NetException("SocketWrapper error in readln "
                                 + e.getMessage());
           }
       }

   /**
     * Reads the next byte array from the socket input stream. This is
     * a blocking read -- it waits indefinitely for the next byte array.
     * <p> Note that the function ignores null array.
     * @return  the value of the next byte packet as a byte array.
     * @exception IOException if there is a communications failure.
     */
   public synchronized byte[] readBytes() throws IOException
       {
       try{
          ensureOpen();
          return getReaderStream().readBytes();
          }
       catch(IOException e)
           {
           this.onException(e);
           this.close();
           throw new NetException("SocketWrapper error in readBytes "
                                 + e.getMessage());
           }
       }

   public InetAddress getInetAddress()
      {return (_socket != null) ?
            _socket.getInetAddress() :
            null;}

   public boolean isOpen()
       {
       return _socket != null;
       }

   public synchronized void flush() throws IOException
       {
       getWriterStream().flush();
       }
       
   public synchronized void close() 
       {
       // make certain this has not already been closed
       if(_socket == null)
          return;

       if(_writer != null)
          {
          try{_writer.close();}
          catch(IOException ignore){}
          finally{_writer = null;}
          }
       if(_reader != null)
          {
          try{_reader.close();}
          catch(IOException ignore){}
          finally{_reader = null;}
          }

       if(_socket != null)
          {
          try{_socket.close();}
          catch(Exception x){}
          finally{_socket = null;}
          }
       }

   private void ensureOpen() throws IOException
      {
      if(_socket == null)
         throw new NetException("Socket is closed");
      }

   private ReaderStream createSocketReader(InputStream in) throws IOException
       {
       if (_binarymode)
          return new PacketInputStream(in);
       else
          return new StringInputStream(in);
       }
       
   private WriterStream createSocketWriter(OutputStream out) throws IOException
       {
       if (_binarymode)
          return new PacketOutputStream(out);
       else
          return new StringOutputStream(out);
       }

   protected void finalize() throws Throwable
      {
      close();
      }
   }
