/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */



package osmq.pulsar;
  
import java.net.*;
import java.io.*;
import osmq.mcns.*;
import osmq.util.*;
import osmq.util.Threaded;
import osmq.records.*;
//import osmq.logs.DateStamp;
/**
  *
  * MCNS (multicast Name Service)
  * MulticastPulsar uses an ip multicast channel to repeatedly broadcast a service name, ip adddress and port 
  * at a fixed time frequency. This feature permits the discovery of a service without having to maintain
  * that information on an NIS or DNS table, and further provides the advantage of not requiring that a 
  * service be tied to a specific server. (Useful for implementing redundant hot backup servers).
  *
  * <p>Another use of the pulsar is to identify whether a service has failed. For example, after a client
  * discovers and attaches to a service, it can listen to the multicast port for the continued broadcast
  * and take appropriate action if and when the broadcast of that service is no longer heard.
  * Similarly, in the event of a server failure, the broadcast channel can be used by a backup server
  * to broadcast a changed ip address / port combination for the service.
  *
  * <p>Reciprocal classes include
  * <p>1) DiscoveryServer -- a service class that performs the function of a name server of MulticastPulsars.
  *                       Clients pass the name of a sercie to the DiscoveryServer and are returned the
  *                       current ip address port number for that service.
  *     It logs 
  *    all active service pulsars and acts as a centralized lookup for clients, and 
  * 2) the ServicePulsarMonitor class which a client uses to monitor the that a service 
  *    remains alive during an active session. 
  * @see osmq.net.DiscoveryServer
  */

public class ServicePulsar extends Threaded
  {
  protected DatagramSocket _socket = null;
  private DataRecord       _mb = new DataRecordBean(4);
  private int              _pulsarPort = MCNS.PULSAR_CHANNEL;
  private int              _delayMillis = Threshold.TEST_ALIVE_FREQUENCY
                              * MCNS.SECONDS;
  private String           _groupaddressString;
  private boolean          isheartbeat = false;

  protected ServicePulsar()
     {
     _groupaddressString = NetworkScope.getScope().getGroupAddress();
     }

  public void setHeartbeat(boolean v)
     {isheartbeat = v;}
  public boolean isHeartbeat()
     {return isheartbeat;}
     
  public void setDiscovery(boolean v)
      {isheartbeat = !v;}
  public boolean isDiscovery()
      {return !isheartbeat;}

  private InetAddress groupaddress;
  byte[] buf = new byte[256];

  private int serviceport = 3000;
  private String servicename;

  DatagramPacket packet;
  Thread t;
  private int priority = Thread.NORM_PRIORITY;


  public void setServicePort(int p)
     {serviceport = p;}
  public int getServicePort()
     {return serviceport;}

  public void setGroupAddress(String a)
     {_groupaddressString = a;}
  public String getGroupAddress()
     {return _groupaddressString;}
     
  public void setServiceName(String s)
     {servicename = s;}
  public String getServiceName()
     {return servicename;}

  protected void iterativeProcess() throws Exception
     {
     _socket.send(packet);
     Thread.sleep(_delayMillis);
     }

  protected void initializer() throws Exception
     {
     buf = servicename.getBytes();
     _pulsarPort = (isheartbeat) ?
                   serviceport :
                   MCNS.PULSAR_CHANNEL;
     _delayMillis = (isheartbeat) ?
              Threshold.HEARTBEAT_FREQUENCY
              * MCNS.SECONDS :
              Threshold.TEST_ALIVE_FREQUENCY
              * MCNS.SECONDS;

     if(isDiscovery())
        _socket = new DatagramSocket(serviceport);
     else
        _socket = new DatagramSocket();
     groupaddress = InetAddress.getByName(_groupaddressString);
     packet = new DatagramPacket(buf, buf.length, groupaddress, _pulsarPort);

     System.out.println(DateStamp.getTimeStamp() +
                           " Multicasting " +
                            new String(buf).trim() +
                            " to " +
                            _groupaddressString +
                            " on port " +
                            _pulsarPort +
                            " every " +
                            (_delayMillis / MCNS.SECONDS) +
                            " secs.");
     }

     
   public DataRecord getRecord()
    {
    try{
       _mb.clearAll();
       _mb.setString(1, servicename);
       _mb.setString(2, InetAddress.getLocalHost().getHostAddress());
       _mb.setInt(3, serviceport);
       return _mb;
       }
    catch(Exception e){e.printStackTrace();}
    return null;
    }

  protected void cleanup()
    {
    if(_socket != null)
       _socket.close();
    _socket = null;
    System.out.println(DateStamp.getTimeStamp() +
                           " Terminating multicaat " +
                            new String(buf).trim() +
                            " to " +
                            _groupaddressString +
                            " on port " +
                            _pulsarPort);
    }


} 
