/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */

package osmq.records;

import java.util.*;
import java.io.*;
import osmq.util.DelimitedStringParser;

/**
  * An implementation of a multi-object container that uses an array
  * to store its values. The container's value can be set individually
  * or using a delimited String, and a delimited String can be generated
  * that represents the many values.
  */

public class CoreRecord implements Cloneable, Serializable
  {

  final public static String    DELIMITER_STRING = "|";
  final protected static int    INIT_CAPACITY = 1024;
  private DelimitedStringParser _parser = new DelimitedStringParser();
  private String                _text = null;
  private Object                _lastfetched = null;
  protected Object[]            _array       = null;
  int                           _size        = 0;

  public CoreRecord()
    {}

  /**
   * Sets the element values from those contained in a delimited string.
   * Parsing of the string is deferred until an element is referrenced.
   */
  public void setFromDelimitedString(String value)
     {
     _text = value;
     _array = null;
     //parsing the string is deferred until an element is referenced
     }

  /**
    * Returns the element values as a delimited string.
    */
  public String getAsDelimitedString()
     {
     if(_text == null)
        _text = mergeToString();

     return _text;
     }

  /**
     * Returns true if the last get() returned a null object, or
     * the last get() object had a string length of zero
     */
  public boolean wasNull()
     {
     return null == _lastfetched ||
        _lastfetched.toString().length() == 0;
     }
  /**
   * Returns an element in the record.
   */
  public Object get(int p)
     {
     parseIfNullArray();
     if(_array == null)
        throw new IllegalStateException("Container size not set");
     _lastfetched = _array[p];
     return _lastfetched;
     }

  /**
   * Sets a record element to the value of an object.
   */
  public void set(int p, Object o)
     {
     if(_array == null)
        throw new IllegalStateException("Container size not set");
     _array[p] = o;
     _text = null;
     }

  /**
   * Resets the number of elements in the record, and clears
   * the value of each element.
   */
  public void resetSize(int size)
     {
     if(size < 0)
        throw new IllegalArgumentException("Illegal size");
     _size = size;
     _array = (_size > 0) ? new Object[size] : null;
     }

  public int size()
     {
     parseIfNullArray();
     return _size;
     }

  /**
   * Sets all elements from an array of object
   */
  public void setAll(Object[] elements)
     {
     _size = elements.length;
     _array = new Object[_size];
     System.arraycopy(elements, 0, _array, 0, _size);
     _text = null;
     }

  /**
   *  Returns all elements as an array
   */
  public Object[] getAll()
     {
     parseIfNullArray();

     // Allocate a target array
     Object[] result = new Object[_size];

     // Copy internal array to target
     if(_size > 0)
        System.arraycopy(_array, 0, result, 0, _size);

     // Return the target
     return result;
     }

   private String mergeToString()
      {
      int factor = 0;
      StringBuffer sb = new StringBuffer(INIT_CAPACITY);

      for(int ctr = 0; ctr < size(); ++ ctr)
           {
           // each element is preceded by a delimiter --
           // even empty elements
           sb.append(DELIMITER_STRING);

           // if there is an element value, append it
           if(get(ctr + factor) != null)
              sb.append(get(ctr).toString());
           }
        return new String(sb);
      }


   private void parseIfNullArray()
     {
     if(_array == null && _text != null)
        parseFromString(_text);
     }

   private void parseFromString(String text)
     {
     // test for valid text string
     if(text == null)
        throw new NullPointerException("invalid text in ExtendedArrayContainer");
     if(text.length() < 1)
        throw new IllegalArgumentException("Invalid text value");

     int nextPosn = 0;
     // reset the parser's text
     if(_parser == null)
        _parser = new DelimitedStringParser();
     _parser.setString(text);

     // reset the DataRecord's container size
     resetSize(_parser.getCount());

     // place a token in each container slot
     while(_parser.hasMoreTokens())
        set(nextPosn++, _parser.nextToken());
     }

  public Object clone()
     {
     try{
        CoreRecord ac = (CoreRecord) super.clone();
        if(_array != null && _size > 0)
           {
           ac._array = this.getAll();
           }
        return ac;
        }
     catch(CloneNotSupportedException e)
        {
        return this;
        }
     }
  }