/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */

package osmq.records;

import java.math.BigDecimal;
import java.util.Date;

import osmq.util.Immutable;
import osmq.util.ImmutableException;
import osmq.util.SpecialChars;
import osmq.util.ByteArray;
import osmq.datarows.DataRow;

/**
  * A multi-element container that sets and returns its elements as
  * either Strings or scalar data types. The DataRecord supports
  * mapping data as an int, double, long, char, String, and Date.
  * The elements are referenced positionally (1-n).  (Note that the indexing scheme
  * can be changed by calling the method setOffsetIndex(int).)
  * The getElementCount() method returns the number of elements
  * in the container.
  * This interface is modeled after the getXXX and setXXX methods implemented by the
  * JDBC ResultSet and Statement classes.
  */

public interface DataRecord extends DataRow, Immutable, ByteArray
  {
  final public static char DEFAULT_DELIMITER = SpecialChars.PIPE_CHAR;


  // --------------- METHODS INHERITED FROM DATA ROW -----------------------------------
  /**
    * Determines whether a position has a value (versus null)
    *
    * @param p the position containing the value to be tested
    * @return  <tt>true</tt> of there is a non-null value at position p, else <i>false</i>
    * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
    */
   boolean hasAValue(int n);


  /**
    * Determines if a position has no value (is empty). Alias method for isNull(int)
    *
    * @param p the position containing the value to be tested
    * @return  <tt>true</tt> if there is a null value (no value) at position p, else <tt>false</tt>
    * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
    */
   boolean isEmpty(int n);


  /**
    * Determines if a position has no value (is empty).  Alias method for isEmpty(int)
    *
    * @param p the position containing the value to be tested
    * @return  <tt>true</tt> if there is a null value (no value) at position p, else <tt>false</tt>
    * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
    */
   boolean isNull(int n);

   /**
     * Returns the value of the designated position.
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the value as an Object or null  if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     */
   Object getObject(int positionIndex);

   /**
     * Returns the value of the designated position as a String
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the position value as a String or a null String if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */
   String getString(int positionIndex);

  /**
    *  Returns the value at the designated position as a String
    *  with all characters forced into upper case.
    *
    * @param positionIndex the first position is 1, the second is 2, ...
    * @return  the value as a String or a null if the position is empty.
    * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid element position.
    */
   String getUpperCaseString(int positionIndex);

  /**
     *  Returns the value at the designated position as a boolean
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the value as a boolean or a null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */
   boolean getBoolean(int positionIndex);

   /**
     * Returns the value at the designated position as a char
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the value as a char or 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */
   char getChar(int positionIndex);

   /**
     * Returns the value at the designated position as a byte
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the value as a byte or 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */
   byte getByte(int positionIndex);

   /**
     * Returns the value at the designated position as an integer
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the value as an int or a 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */

   int getInt(int positionIndex);

   /**
     *  Returns the value at the designated position as a short integer
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the position value as a short or 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if columnIndex is not a valid element position.
     */
   short getShort(int positionIndex);


   /**
     * Returns the value at the designated position as a long integer
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the position value as a long or 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */
   long getLong(int positionIndex);

   /**
     * Returns the value at the designated position as a double
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the position value as a double or 0.0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     */
   double getDouble(int positionIndex);

   /**
     *  Returns the value at the designated position as a float
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the position value as a double or 0.0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     */
   float getFloat(int positionIndex);

  /**
     *  Returns the value at the designated position as a BigDecimal
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the position value as a BigDecimal or null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     */
   BigDecimal getBigDecimal(int positionIndex);

   /**
     *  Returns the value at the designated position as a Date
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the position value as a Date or null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     */
   Date getDate(int positionIndex);


 /**
     * Returns the value at the designated position
     * as a formatted date string in CCYY-MM-DD format.
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the position value as a String or null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     */
   String getFormattedDateString(int posn);

   /**
     * Returns the value at the designated position
     * as an unformatted date string in CCYYMMDD format.  (This is the normal
     * physical storage format for a date).
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the position value as a String or null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     * @see #setDateString(int, String)
     */
   String getUnformattedDateString(int posn);

   /**
     * Returns the value at the designated position as a java.SQL.Date
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the position value as an sql Date or null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     * @see #setDateString(int, String)
     */
   java.sql.Date getSQLDate(int posn);

   /**
     * Returns the base index currently used to reference elements.
     */
   int getOffsetIndex();

  /**
    * Returns the number of element positions (including those that are empty)
    *
    * @return  the number of DataRecord elements
    * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
    */
   int getElementCount();

   // -------------------------- SETTERS -------------------------------------------

   /**
     * Clears the value at the designated position, setting it to null (or zero if numeric)
     * @exception ImmutableDataException if the DataRecord object is immutable.
     */
   void clear(int positionIndex);

   /**
     * Sets the value at the designated position
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if p is not a valid element position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
   void setObject(int positionIndex, Object value) throws ImmutableException;

   /**
     * Sets the value at the designated position to null
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @exception  ArrayIndexOutOfBoundsException if p is not a valid element position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
   void setNull(int positionIndex) throws ImmutableException;

   /**
     * Sets the element at the designated position to a String value
     *
     * @param p the element's relative position (1 to n)
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if p is not a valid element position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
   void setString(int positionIndex, String value) throws ImmutableException;

   /**
     * Sets the element at the designated position to a String value.  Setting the value also
     * eliminates any leading or trailing spaces.
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
   void setTrimmedString(int positionIndex, String value) throws ImmutableException;

   /**
     * Sets the element at the designated position to a String value that represents a decimal
     * value.   Setting the element ensures that the string has a decimal component,
     * and excessive non-significant decimal values (trailing zeros) are dropped.
     * A value of "5" will result in a string value of "5.0" and a value of "5.400"
     * will result in a string value of "5.4".
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
   void setDecimalString(int posn, String value) throws ImmutableException;


   /**
     * Sets the element at the designated position to a boolean value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
   void setBoolean(int positionIndex, boolean value) throws ImmutableException;

   /**
     * Sets the element at the designated position to an
     * upper-case String value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
   void setUpperCaseString(int positionIndex, String value) throws ImmutableException;

   /**
     * Sets the element at the designated position to an int value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
   void setInt(int positionIndex, int value) throws ImmutableException;

   /**
     * Sets the element at the designated position to a char value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
   void setChar(int positionIndex, char value) throws ImmutableException;

    /**
     * Sets the element at the designated position to a byte value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
   void setByte(int positionIndex, byte value) throws ImmutableException;

   /**
     * Sets the element at the designated position to a short value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
   void setShort(int positionIndex, short value) throws ImmutableException;

   /**
     * Sets the element at the designated position to a long value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
   void setLong(int positionIndex, long value) throws ImmutableException;

   /**
     * Sets the element at the designated position to a double value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid  position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
   void setDouble(int positionIndex, double value) throws ImmutableException;

   /**
     * Sets the element at the designated position to a float value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid  position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
   void setFloat(int positionIndex, float value) throws ImmutableException;

   /**
     * Sets the element at the designated position to a Date value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid  position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
   void setDate(int positionIndex, Date value) throws ImmutableException;

   void setDateString(int posn, String value) throws ImmutableException;

   /**
     * Sets the element at the designated position to a BigDecimal value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid  position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
   void setBigDecimal(int positionIndex, BigDecimal value) throws ImmutableException;

   /**
     * Sets the base index used to reference elements.
     * If not set explicitly, this index defaults to 1.
     * @param value the new base index value
     */
   void setOffsetIndex(int value);

   /**
     * Returns the character that is used to delimit the various elements when the elements
     * are strung together.  The string of elements is accessible by calling the getText()
     * method.
     */
  char getDefaultDelimiter();


   /**
     * Returns the elements as an array of bytes. The first
     * byte in the array is the element delimiter.
     */
   byte[] getBytes();

   /**
     * Creates a delimited String from the elements
     * The first character in the String is assumed to be the delimiting character.
     */
   String getText();

   /**
     * Sets the elements to the values found in a delimited String.
     * The first character in the String is assumed to be the delimiting character.
           * @exception ImmutableDataException if the DataRecord object is immutable.
     */
   void setText(String value);

   /**
     * Sets the elements to the values found in a delimited array of bytes.
     * The first byte in the array is assumed to be the delimiting character.
     * @exception ImmutableDataException if the DataRecord object is immutable.
     */
   void setBytes(byte[] value);

   /**
     *  Returns an array that contains all element values
     */
   Object[] getAsArray();


   /**
     * Resets the size (number of elements), and clears all the values
     * @param size the number of elements in the resized object
     * @exception ImmutableDataException if the DataRecord object is immutable
     */
   void resetSize(int size) throws ImmutableException;


   /**
     * Clears all the elements, setting them to null (or zero if numeric)
     * @exception ImmutableDataException if the DataRecord object is immutable.
     */
   void clearAll() throws ImmutableException;


   /**
     * Sets the data elements from the values in an array.
     * @exception ImmutableException if the data elements are immutable.
     */
   void setFromArray(Object[] v) throws ImmutableException;


   /** Returns <tt>true</tt> if the <tt>last fetched</tt> element was a null object, or
     * the <tt>last fetched</tt> has a string length of zero.  (The <tt>last fetched</tt>
     * is the element referenced by the last <code>getXXX()</code> method call.)
     */
   boolean wasNull();

   Object clone();

   }



