/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */

package osmq.records;

import java.util.*;
import java.math.*;

import java.io.Serializable;
import java.lang.System;
import java.util.*;
import java.math.BigDecimal;
import java.util.Date;
import java.text.DateFormat;

import osmq.util.DelimitedStringParser;
import osmq.util.Immutable;
import osmq.util.ImmutableException;
import osmq.util.ImmutableImpl;
import osmq.util.DecimalConverter;
import osmq.util.Formatter;
import osmq.util.SpecialChars;
import osmq.datarows.DataRow;

 /**
    * Default implementation of a DataRecord. Uses a CoreRecord
    * as an aggregate object container.
    */
public  class DataRecordBean extends ImmutableImpl
                                implements DataRecord, DataRow, Cloneable, Serializable
  {
  private int                   _indexfactor = 1;
  private CoreRecord            _container;
  private DecimalConverter      _converter = null;
  private Calendar              _calendar = null;


  public DataRecordBean()
     {
     this(new CoreRecord());
     }

  public DataRecordBean(Object[] elements)
     {
     this();
     setFromArray(elements);
     }

  public DataRecordBean(int sz)
     {
     this();
     resetSize(sz);
     }

  public DataRecordBean(CoreRecord c)
     {
     _container = c;
     }

  public DataRecordBean(String value)
     {
     this();
     setText(value);
     }

  public DataRecordBean(byte[] value)
     {
     this();
     setBytes(value);
     }

   public  char getDefaultDelimiter()
     {
     return DEFAULT_DELIMITER;
     }

  public boolean hasADelimiter()
     {
     return true;
     }

   /**
     * Clear all the elements in the DataRecord object to null (or zero if numeric)
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void clearAll()  throws ImmutableException
     {
     throwIfImmutable();
     resetSize(getElementCount());
     }

   /**
     * Clears the value of the designated position to null (or zero if numeric)
     * @exception ImmutableException if the DataRecord object is immutable.
     * @see #clearAll()
     */
  public void clear(int posn) throws ImmutableException
     {
     setNull(posn);
     }

  // -------------------------------------------------------------------------
  // -------------------- UTILITY --------------------------------------------
  // -------------------------------------------------------------------------
  public DataRecord getClone()
     {
     return (DataRecord) this.clone();
     }

  public String toString()
     {
     return getText();
     }


  /**
    * Sets the indexing scheme by indicating the index of the
    * first element in the DataRecord.
    * <p> Typical values:
    * <p> 0 indicates a scheme of 0 thru n-0,
    * <p> 1 indicates a scheme of 1 thru n.
    */
  public void setOffsetIndex(int value)
     {
     _indexfactor = value;
     }

   /**
     * Returns the base index currently used to reference elements in the DataRecord.
     */
   public int getOffsetIndex()
     {
     return _indexfactor;
     }

   /** Returns <tt>true</tt> if the <tt>lastfetched</tt> element was a null object, or
     * an object with a string length of zero.  (The <tt>last fetched</tt>
     * is the element referenced by the last <code>getXXX()</code> method call.)
     */
  public boolean wasNull()
     {
     return _container.wasNull();
     }

   /**
     * Creates a delimited String from the DataRecord elements.
     * The first character in the String is the delimiting character.
     */
  public String getText()
     {
     /*
      * When the _text string is set, subsequent calls to getText() will return
      * that string without building it from the many elements. However, setting any
      * element in the record results int _text being set to null, and getText() must
      * reassemble it.
      */
     return _container.getAsDelimitedString();
     }

  /**
     * Returns the dataset elements as an array of bytes.
     * <p>The first byte in the array is the element delimiter.
     */
  public byte[] getBytes()
     {
     return getText().getBytes();
     }

  /**
     *  Gets the value at the designated position in the DataRecord object as an Object
     *
     * @param posn the relative position, based on the indexing scheme
     * @return  the position value as a char, or '\0' if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     */
  public Object getObject(int posn)
     {
     try{
        return getBaseRecord().get(posn - _indexfactor);
        }
     catch(ArrayIndexOutOfBoundsException e)
        {
        System.out.println("Index out of bounds in DefaultDataRecord");
        e.printStackTrace();
        throw e;
        }
      }

   /**
     *  Returns an array that contains all element values in the DataRecord
     */
  public Object[] getAsArray()
     {
     return getBaseRecord().getAll();
     }

   /**
     *  Gets the value at the designated position in the DataRecord object as a char
     *
     * @param posn the relative position, based on the indexing scheme
     * @return  the position value as a char, or '\0' if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
           *             valid position.
     */
  public char getChar(int posn)
     {
     try{return getString(posn).charAt(0);}
     catch(StringIndexOutOfBoundsException e) {return SpecialChars.NULL_CHAR;}
     }

  /**
     *  Gets the value at the designated position in the DataRecord object as a byte
     *
     * @param posn the relative position, based on the indexing scheme
     * @return  the position value as a byte, or binary 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     */
  public byte getByte(int posn)
     {
     return (byte) getChar(posn);
     }

   /**
     *  Gets the value at the designated position in the DataRecord object as a boolean
     *
     * @param posn the relative position, based on the indexing scheme
     * @return  the position value as a boolean (<i>true</i> or <i>false</i>),
     *          or false if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     */
  public boolean getBoolean(int posn)
     {
     try{return (getString(posn).charAt(0) == 'T') ? true : false;}
     catch(StringIndexOutOfBoundsException e) {return false;}
     }

   /**
     *  Gets the value at the designated position in the DataRecord object as a float
     *
     * @param posn the relative position, based on the indexing scheme
     * @return  the position value as a float, or 0.0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     */
  public float getFloat(int posn)
     {
     try{return Float.parseFloat(getString(posn));}
     catch(NumberFormatException e){return (float) 0;}
     }

   /**
     *  Gets the value at the designated position in the DataRecord object as a double
     *
     * @param posn the relative position, based on the indexing scheme
     * @return  the position value as a double, or 0.0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     */
  public double getDouble(int p)
     {
     try{return Double.parseDouble(getString(p));}
     catch(NumberFormatException e){return 0.0;}
     }

  /**
     *  Gets the value at the designated position in the DataRecord object as a BigDecimal
     *
     * @param posn the relative position, based on the indexing scheme
     * @return  the position value as a BigDecimal, or null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     */
  public BigDecimal getBigDecimal(int p)
     {
     try{return new BigDecimal(getString(p));}
     catch(NumberFormatException e){return null;}
     }

  public int getElementCount()
     {
     return getBaseRecord().size();
     }

  // -------------------- POSITIONAL GETTERS ----------------------
   /**
     *  Gets the value at the designated position in the DataRecord object as an integer
     *
     * @param posn the relative position, based on the indexing scheme
     * @return  the position value as an int, or 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     */
  public int getInt(int p)
     {
     try{return Integer.parseInt(getString(p));}
     catch(NumberFormatException e) {return 0;}
     }

   /**
     *  Gets the value at the designated position in the DataRecord object as a long
     *
     * @param posn the relative position, based on the indexing scheme
     * @return  the position value as a long, or 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     */
  public long getLong(int p)
     {
     try{return Long.parseLong(getString(p));}
     catch(NumberFormatException e){return 0;}
     }

   /**
     *  Gets the value at the designated position in the DataRecord object as a short
     *
     * @param posn the relative position, based on the indexing scheme
     * @return  the position value as a short, or 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     */
  public short getShort(int n)
     {
     return (short) getInt(n);
     }

   /**
     *  Gets the value at the designated position in the DataRecord object as a String
     *
     * @param posn the relative position, based on the indexing scheme
     * @return  the position value as a String or null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     */
  public String getString(int posn)
     {
     Object o = getObject(posn);
     return (o != null) ? o.toString() : SpecialChars.NULL_STRING;
     }

  public String getUpperCaseString(int n)
     {
     return getString(n).toUpperCase();
     }

   /**
     *  Gets the value at the designated position in the DataRecord object as a Date
     *
     * @param posn the relative position, based on the indexing scheme
     * @return  the position value as a Date or null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     */
  public Date getDate(int posn)
     {
     return  Formatter.stringToDate(getUnformattedDateString(posn));
     }

  public boolean hasAValue(int posn)
     {
     try{
        return getString(posn).length() > 0;
        }
     catch(Exception e)
        {
        return false;
        }
     }

  public boolean isEmpty(int posn)
     {
     return isNull(posn);
     }

   public boolean isNull(int posn)
     {
     return !hasAValue(posn);
     }

   /**
     * Resets the size (number of elements) in the DataRecord object, and clears all the values
     * @param size the number of elements in the resized DataRecord object
     * @exception ImmutableException if the DataRecord object is immutable
     */
   public void resetSize(int newsize) throws ImmutableException
      {
      throwIfImmutable();
     // _text = null;
      getBaseRecord().resetSize(newsize);
      }

  //-------------------- SETTERS ---------------------

   /**
     * Sets the element at the designated position in the DataRecord object to a Object value
     *
     * @param posn the relative position of the element, based on the indexing scheme
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if posn is not a valid  position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setObject(int posn, Object value)  throws ImmutableException
      {
      this.throwIfImmutable();
      try{
         getBaseRecord().set(posn - _indexfactor, value);
         }
      catch(ArrayIndexOutOfBoundsException e)
         {
         System.out.println("Index out of bounds in DefaultDataRecord");
           e.printStackTrace();
         throw e;
         }
      }

   /**
     * Sets the value of element in the DataRecord to values found in a delimited String.
     * @param text the delimited String of elements.  The first character in the String
     * is assumed to be the delimiter.
     * @exception ImmutableException if the DataRecord object is immutable
     */
   public void setText(String text)
      {
      throwIfImmutable();
      _container.setFromDelimitedString(text);
      }

   /**
     * Sets the value of element in the DataRecord to values found in a byte array
     * @param text the array of bytes.  The first byte in the array
     * is assumed to be the delimiter.
     * @exception ImmutableException if the DataRecord object is immutable
     */
  public void setBytes(byte[] value)
     {
     setText(new String(value));
     }

   /**
     * Sets the element at the designated position in the DataRecord object to a String value
     *
     * @param posn the relative position of the element, based on the indexing scheme
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if posn is not a valid  position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setString(int posn, String s)  throws ImmutableException
     {
     setObject(posn, s);
     }

  public void setTrimmedString(int posn, String s)  throws ImmutableException
     {
       if(s != null && s.length() > 0)
               setString(posn, s.trim());
       else
           setString(posn, s);
      }

   /**
     * Sets the element at the designated position in the DataRecord object to null
     *
     * @param posn the relative position of the element, based on the indexing scheme
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if posn is not a valid  position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setNull(int posn) throws ImmutableException
     {
     setObject(posn, null);
     }

   /**
     * Sets the element of the designated position in the DataRecord object to a byte value
     *
     * @param posn the relative position of the element, based on the indexing scheme
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if posn is not a valid  position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setByte(int posn, byte value)  throws ImmutableException
           {
           setString(posn, new String("" + value));
           }

   /**
     * Sets the element of the designated position in the DataRecord object to a boolean value
     *
     * @param posn the relative position of the element, based on the indexing scheme
     * @param value the element's new value, either <i>true</i> or <i>false</i>.
     * @exception  ArrayIndexOutOfBoundsException if posn is not a valid  position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setBoolean(int posn, boolean value)  throws ImmutableException
     {
     setString(posn, new String((value) ? "T" : "F"));
     }

   /**
     * Sets the element of the designated position in the DataRecord object to a char value
     *
     * @param posn the relative position of the element, based on the indexing scheme
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if posn is not a valid  position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setChar(int posn, char value)  throws ImmutableException
     {
     setString(posn, new String("" + value));
     }

   /**
     * Sets the element of the designated position in the DataRecord object to a float value
     *
     * @param posn the element's relative position, based on the indexing scheme
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if posn is not a valid  position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
   public void setFloat(int posn, float value)  throws ImmutableException
     {
     setString(posn, String.valueOf(value));
     }

   /**
     * Sets the element of the designated position in the DataRecord object to a double value
     *
     * @param posn the element's relative position, based on the indexing scheme
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if posn is not a valid  position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setDouble(int posn, double value)  throws ImmutableException
     {
     setString(posn, getConverter().setAndTranslate(value));
     }

  public void setDecimalString(int posn, String value) throws ImmutableException
      {
      setString(posn, Formatter.formatDecimals(value));
      }

  /**
     * Sets the element of the designated position in the DataRecord object to a BigDecimal value
     *
     * @param posn the element's relative position, based on the indexing scheme
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if posn is not a valid  position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setBigDecimal(int posn, BigDecimal value) throws ImmutableException
     {
     setString(posn, value.toString());
     }



   /**
     * Sets the element of the designated position in the DataRecord object to a date/time value
     *
     * @param posn the element's relative position, based on the indexing scheme
     * @param value the element's new value as a Java Date object
     * @exception  ArrayIndexOutOfBoundsException if posn is not a valid  position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setDate(int posn, Date value) throws ImmutableException
     {
     setString(posn, Formatter.javaDateToString(value));
     }


  public void setDateString(int posn, String value) throws ImmutableException
     {
     setString(posn, Formatter.unformatDate(value));
     }

  public String getUnformattedDateString(int posn)
     {
     return getString(posn);
     }

  public String getFormattedDateString(int posn)
     {
     return isEmpty(posn) ?
            null : Formatter.formatDateString(getString(posn));
     }

  public java.sql.Date getSQLDate(int posn)
     {
     return isEmpty(posn) ?
            null :  java.sql.Date.valueOf(getFormattedDateString(posn));

     }

  public void setFromArray(Object[] s) throws ImmutableException
     {
     getBaseRecord().setAll(s);
     }

   /**
     * Sets the element of the designated position in the DataRecord object to an integer value
     *
     * @param posn the element's relative position, based on the indexing scheme
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if posn is not a valid  position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setInt(int posn, int value)  throws ImmutableException
     {
     setString(posn, String.valueOf(value));
     }

   /**
     * Sets the element of the designated position in the DataRecord object to a long value
     *
     * @param posn the element's relative position, based on the indexing scheme
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if posn is not a valid  position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setLong(int posn, long value)  throws ImmutableException
     {
     setString(posn, String.valueOf(value));
     }

   /**
     * Sets the element of the designated position in the DataRecord object to a short value
     *
     * @param posn the element's relative position, based on the indexing scheme
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if posn is not a valid  position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setShort(int posn, short value)  throws ImmutableException
     {
     setString(posn, String.valueOf(value));
     }

  public void setUpperCaseString(int posn, String s)  throws ImmutableException
     {
     setObject(posn, s.toUpperCase());
     }

  protected CoreRecord getBaseRecord()
     {
     return _container;
     }

  private DecimalConverter getConverter()
     {
     if(_converter == null)
        _converter = new DecimalConverter();
     return _converter;
     }

  private Calendar getCalendar()
     {
     if(_calendar == null)
        _calendar = Calendar.getInstance();
     return _calendar;
     }

   public Object clone()
     {
     if(this.isImmutable())
        return this;
     DataRecordBean dds = (DataRecordBean) super.clone();
     dds._container = (CoreRecord) this._container.clone();
     dds._indexfactor = _indexfactor;
     return dds;
     }

  public static void main(String[] vars)
     {
     DataRecordBean dds = new DataRecordBean("|1|2|3");
     System.out.println(dds.getText());
     int i = dds.getInt(1);
     System.out.println("Text = " + dds.getText());
     dds.setInt(2, 55);
     System.out.println("Text = " + dds.getText());
     }
  }
