/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */


package osmq.records;

import java.util.*;
import java.math.*;
import java.io.Serializable;
import java.lang.System;
import java.util.*;


public  class DataRecordWrapper implements DataRecord, Cloneable, Serializable
   {
   private DataRecord _ds;
   public DataRecordWrapper(DataRecord ds)
      {
      _ds = ds;
      }

  /**
    * Returns true if this is an immutable message.
    * @see #setImmutable()
    */
   public boolean isImmutable()
      {
      return _ds.isImmutable();
      }

  /**
    * Sets the DataRecord's state to "immutable."  The values of
    * an immutable DataRecord cannot be modified.  Any attempt to
    * modify those values will result in an ImmutableException being
    * thrown.
    * @see osmq.util.ImmutableException
    */
   public void setImmutable()
      {
      _ds.setImmutable();
      }


   /**
     * Clears all the elements in the DataRecord object, setting them to null (or zero if numeric)
     * @exception ImmutableDataException if the DataRecord object is immutable.
     */
   public void clearAll()  //throws ImmutableDataException
      {
      _ds.clearAll();
      }

   /**
     * Clears the value at the designated position, setting it to null (or zero if numeric)
     * @exception ImmutableDataException if the DataRecord object is immutable.
     * @see #clearAll()
     */
   public void clear(int p) //throws ImmutableDataException
      {
      _ds.clear(p);
      }

   public String toString()
      {
      return _ds.toString();
      }

   /**
     * Creates a delimited String from the DataRecord elements
     * The first character in the String is assumed to be the delimiting character.
     */
   public String getText()
      {
      return _ds.getText();
      }

   /**
     * Returns the dataset elements as an array of bytes. The first
     * byte in the array is the element delimiter.
     */
   public byte[] getBytes()
      {
      return _ds.getBytes();
      }

   /**
     * Returns the value of the designated position in the DataRecord.
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the value as an Object or null  if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
           *             valid position.
     */
   public Object getObject(int p)
      {
      return _ds.getObject(p);
      }

   public Object[] getAsArray()
      {
      return _ds.getAsArray();
      }

   /**
     * Returns the value at the designated position in the DataRecord object as a char
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the value as a char or 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */
   public char getChar(int p)
      {
      return _ds.getChar(p);
      }


   /**
     * Returns the value at the designated position in the DataRecord object as a byte
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the value as a byte or 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */
  public byte getByte(int p)
     {
     return _ds.getByte(p);
     }

  /**
     *  Returns the value at the designated position in the DataRecord object as a boolean
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the value as a boolean or a null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */
  public boolean getBoolean(int p)
     {
     return _ds.getBoolean(p);
     }

   /**
     *  Returns the value at the designated position in the DataRecord object as a float
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the position value as a double or 0.0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
         *             valid position.
     */
   public float getFloat(int p)
      {
       return _ds.getFloat(p);
       }

   /**
     * Returns the value at the designated position in the DataRecord object as a double
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the position value as a double or 0.0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     */
   public double getDouble(int p)
      {
      return _ds.getDouble(p);
      }

   /**
     *  Returns the value at the designated position in the DataRecord object as a BigDecimal
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the position value as a BigDecimal or null if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a
     *             valid position.
     */
   public BigDecimal getBigDecimal(int p)
      {
      return _ds.getBigDecimal(p);
      }

  /**
    * Returns the number of positions in the DataRecord (including those that are empty)
    *
    * @return  the number of DataRecord elements
    * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
    */

   public int getElementCount()
      {
      return _ds.getElementCount();
      }

  // -------------------- POSITIONAL GETTERS ----------------------

   /**
     * Returns the value at the designated position in the DataRecord object as an integer.
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the value as an int or a 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */
   public int getInt(int p)
       {
       return _ds.getInt(p);
       }

   /**
     *  Returns the value at the designated position in the DataRecord object as a long integer
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the position value as a short or 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if columnIndex is not a valid element position.
     */
   public long getLong(int p)
      {
      return _ds.getLong(p);
      }

   /**
     *  Returns the value at the designated position in the DataRecord object as a short integer
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the position value as a short or 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if columnIndex is not a valid element position.
     */
   public short getShort(int p)
      {
      return _ds.getShort(p);
      }

   /**
     * Returns the value of the designated position in the DataRecord object as a String
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the position value as a String or a null String if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     */
   public String getString(int p)
      {
      return _ds.getString(p);
      }

  /**
    *  Returns the value at the designated position in the DataRecord object as a String
          *  with all characters forced into upper case.
    *
    * @param positionIndex the first position is 1, the second is 2, ...
    * @return  the value as a String or a null if the position is empty.
    * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid element position.
    */
   public String getUpperCaseString(int p)
      {
      return _ds.getUpperCaseString(p);
      }

   /**
     *  Returns the value at the designated position in the DataRecord object as a Java Date
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @return  the position value as a short or 0 if the position is empty.
     * @exception  ArrayIndexOutOfBoundsException if columnIndex is not a valid element position.
     */
  public Date getDate(int p)
     {
     return _ds.getDate(p);
     }

  /**
    * Determines whether a position has a value (versus null)
    *
    * @param p the position containing the value to be tested
    * @return  <tt>true</tt> of there is a non-null value at position p, else <i>false</i>
    * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
    */
  public boolean hasAValue(int p)
      {
      return _ds.hasAValue(p);
      }

  /**
    * Determines if a position has no value (is empty).  Alias for isNull(int)
    *
    * @param p the position containing the value to be tested
    * @return  <tt>true</tt> if there is a null value (no value) at position p, else <tt>false</tt>
    * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
    */
  public boolean isEmpty(int p)
     {
     return _ds.isEmpty(p);
     }

    /**
    * Determines if a position has no value (is a null value).  Alias for isEmpty(int)
    *
    * @param p the position containing the value to be tested
    * @return  <tt>true</tt> if there is a null value (no value) at position p, else <tt>false</tt>
    * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
    */
  public boolean isNull(int p)
     {
     return _ds.isNull(p);
     }
     
   /**
     * Resets the size (number of elements) in the DataRecord object, and clears all the values
           * @param size the number of elements in the resized DataRecord object
     * @exception ImmutableDataException if the DataRecord object is immutable
     */
  public void resetSize(int newsize)// throws ImmutableDataException
     {
     _ds.resetSize(newsize);
     }

  //-------------------- SETTERS ---------------------
   /**
     * Sets the DataRecord elements to the values found in a delimited String.
     * The first character in the String is assumed to be the delimiting character.
     * @exception ImmutableDataException if the DataRecord object is immutable.
     */
  public void setText(String value)
     {
     _ds.setText(value);
     }


   /**
     * Sets the DataRecord elements to the values found in a delimited array of bytes.
     * The first byte in the array is assumed to be the delimiting character.
           * @exception ImmutableDataException if the DataRecord object is immutable.
     */
  public void setBytes(byte[] value)
     {
     _ds.setBytes(value);
     }


   /**
     * Sets the value at the designated position in the DataRecord object
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if p is not a valid element position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setObject(int p, Object value)//  throws ImmutableDataException
     {
     _ds.setObject(p, value);
     }

   /**
     * Sets the element at the designated position in the DataRecord object to a String value
     *
     * @param p the element's relative position (1 to n)
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if p is not a valid element position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setString(int p, String s) // throws ImmutableDataException
      {
      _ds.setString(p, s);
      }

   public void setTrimmedString(int p, String s) // throws ImmutableDataException
      {
      _ds.setTrimmedString(p, s);
      }
   /**
     * Sets the value at the designated position in the DataRecord to null
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @exception  ArrayIndexOutOfBoundsException if p is not a valid element position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setNull(int p)// throws ImmutableDataException
    {
    _ds.setNull(p);
    }

   /**
     * Sets the element at the designated position in the DataRecord object to a byte value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setByte(int p, byte value)  //throws ImmutableDataException
     {
     _ds.setByte(p, value);
     }


   /**
     * Sets the element at the designated position in the DataRecord object to a boolean value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setBoolean(int p, boolean value) // throws ImmutableDataException
     {
     _ds.setBoolean(p, value);
     }

   /**
     * Sets the element at the designated position in the DataRecord object to a char value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setChar(int p, char value)  //throws ImmutableDataException
      {
      _ds.setChar(p, value);
      }

   /**
     * Sets the element at the designated position in the DataRecord object to a float value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid  position.
           * @exception ImmutableException if the DataRecord object is immutable.
     */
   public void setFloat(int p, float value)  //throws ImmutableDataException
      {
      _ds.setFloat(p, value);
      }

   /**
     * Sets the element at the designated position in the DataRecord object to a double value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid  position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setDouble(int p, double value)  //throws ImmutableDataException
     {
     _ds.setDouble(p, value);
     }


   /**
     * Sets the element at the designated position in the DataRecord object to a decimal String value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid  position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setDecimalString(int p, String value)
     {
     _ds.setDecimalString(p, value);
     }

   /**
     * Sets the element at the designated position in the DataRecord object to a BigDecimal value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid  position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setBigDecimal(int p, BigDecimal value) //throws ImmutableDataException
    {
    _ds.setBigDecimal(p, value);
    }

    /**
     * Sets the element at the designated position in the DataRecord object to a Date value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid  position.
           * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setDate(int p, Date value) //throws ImmutableDataException
    {
    _ds.setDate(p, value);
    }

   /**
     * Sets the data elements from the values in an array.
     * @exception ImmutableException if the data elements are immutable.
           */
  public void setFromArray(Object[] s)   //throws ImmutableDataException
     {
     _ds.setFromArray(s);
     }


   /**
     * Sets the element at the designated position in the DataRecord object to an int value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
           * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setInt(int p, int value)  //throws ImmutableDataException
     {
     _ds.setInt(p, value);
     }

   /**
     * Sets the element at the designated position in the DataRecord object to a long value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setLong(int p, long value)  //throws ImmutableDataException
          {
          _ds.setLong(p, value);
          }

   /**
     * Sets the element at the designated position in the DataRecord object to a short value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setShort(int p, short value)  //throws ImmutableDataException
      {
      _ds.setShort(p, value);
      }

   /**
     * Sets the element at the designated position in the DataRecord object to an
           * upper-case String value
     *
     * @param positionIndex the first position is 1, the second is 2, ...
     * @param value the element's new value
     * @exception  ArrayIndexOutOfBoundsException if positionIndex is not a valid position.
     * @exception ImmutableException if the DataRecord object is immutable.
     */
  public void setUpperCaseString(int p, String value)  //throws ImmutableDataException
     {
     _ds.setUpperCaseString(p, value);
     }


   /**
     * Sets the base index used to reference elements in the DataRecord.
     * If not set explicitly, this index defaults to 1.
     * @param value the new base index value
         */
  public void setOffsetIndex(int value)
     {
     _ds.setOffsetIndex(value);
     }


   /**
     * Returns the base index currently used to reference elements in the DataRecord.
     * @see #setOffsetIndex(int)
     */
  public int getOffsetIndex()
     {
         return _ds.getOffsetIndex();
     }


   /** Returns <i>true</i> if the <i>lastfetched</i> element was a null object, or
     * an object with a string length of zero.  (The <i>last fetched</i>
     * is the element referenced by the last <code>getXXX()</code> method call.)
     */
  public boolean wasNull()
     {
     return _ds.wasNull();
     }


  public void setDateString(int posn, String value)
     {
     _ds.setDateString(posn, value);
     }

  public String getFormattedDateString(int posn)
     {
     return _ds.getFormattedDateString(posn);
     }


  public String getUnformattedDateString(int posn)
     {
     return _ds.getUnformattedDateString(posn);
     }

  public java.sql.Date getSQLDate(int posn)
     {
     return _ds.getSQLDate(posn);
     }

  public char getDefaultDelimiter()
     {
     return _ds.getDefaultDelimiter();
     }


  public boolean hasADelimiter()
     {
     return _ds.hasADelimiter();
     }


  protected void setDataRecord(DataRecord ds)
     {
     _ds = ds;
     }

  public Object clone()
     {
     if(this.isImmutable())
        return this;
     try{DataRecordWrapper w = (DataRecordWrapper) super.clone();
        w._ds =  (_ds != null) ? (DataRecord) _ds.clone() : null;
        return w;
        }
     catch(CloneNotSupportedException e){return this;}
     }
  }
