/* ====================================================================
 *
 *
 * Copyright (c) 1999-2002 CPortals Systems, Inc.  All rights reserved.
 *
 * This file is part of OSMQ.
 *
 *   OSMQ is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   OSMQ is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with OSMQ; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * OSMQ is a trademark of CPortals Systems, Inc.
 * ====================================================================
 */

package osmq.util;

import java.io.*;
import java.net.*;
import java.util.*;


 /**
   * Base class for objects that need a seperate thread to iteratively perform a task.
   * Handles properly destroying the object when the thread terminates.
   */
public abstract class Threaded implements Runnable
   {
   // Class-level variables
   private static ThreadGroup threadedgroup;
   private static int DEFAULT_YIELD_COUNT = 10;
   
   // For assigning a unique sequence identifier to a class
   private static int threadSeqNumber = 0;

   protected int      _mythreadsequence;       // id for this class instance
   private Object     _threadlock = new Object();  // used for synchronization
   private int        _yieldcount;
   private int        _priority;
   private boolean    _paused = false;
   private boolean    _rancleanup = false;
   
   // flag that initialization completed successfully
   private boolean    _successfulinit = false;

   private boolean    _didstart = false; // start method was called
   private volatile   boolean _shutdowncompleted = false;
   private volatile Thread _thread = null;

   // Tracks the number of times iterativeProcess() is called
   private long       _iterations = 0;

   private ExceptionListener _listener;

   private String     _userdefinedname = "";

   // Set if terminating due to an exception thrown by iterativeProcess()
   private Exception  terminatingexception = null; 

   
   final public static int INCREASED_PRIORITY = Thread.NORM_PRIORITY + 1;
   final public static int HIGH_PRIORITY      = Thread.NORM_PRIORITY + 2;
   final public static int VERY_HIGH_PRIORITY = Thread.NORM_PRIORITY + 3;
   final public static int HIGHEST_PRIORITY   = Thread.NORM_PRIORITY + 4;

    // -------------------------------------------------------------------
   public Threaded()
      {
      _yieldcount = DEFAULT_YIELD_COUNT;
      _priority = Thread.NORM_PRIORITY;
      }

   abstract protected void initializer() throws Exception;

   // To gracefully exit from the thread, call stop() from within
   // iterative process.
   abstract protected void iterativeProcess() throws Exception;

    // finalizer must be able to be called multiple times
   abstract protected void cleanup();

   // -------------------------------------------------------------------

   public void setUserDefinedName(String name)
      {_userdefinedname = name;}
   public String getUserDefinedName()
      {return _userdefinedname;}

   /**
     * This function destroys all the threads derived from
     * Threaded.  It should not be called except in catastrophic
     * situations.
     */
   public static void killAllInGroup()
      {
      getThreadGroup().destroy();
      }
      
   /**
     * Registers a listener that will be notified if the thread is being halted
     * due to an exception that was thrown by the derived class's iterativeProcess()
     * method. 
     *
     */
   public void setExceptionListener(ExceptionListener listener)
      {_listener = listener;}
   public ExceptionListener getExceptionListener()
      {return _listener;}

  public boolean successfulInit(){return _successfulinit;}
   
  /**
    * This method first calls the derived class's initialize() method, and then
    * begins iteratively executing the derived class's iterativeProcess() function.
    * The thread continues calling iterativeProcess() until the the stop() method is
    * invoked, or an exception is thrown by the derived class's interativeProcess() function.
    * <P>If an exception is thrown by the iterativeProcess() method, the exception listener
    * (if one has been registered) will be notified.
    * <P>As a final step when the thread is halted, whether by a stop() call or due to
    * an exception being thrown, the derived class's cleanup() method will be called.
    *
    *
    * @exception  IllegalStateException if the initialize() method fails or
    *             a thread could not be started.
    * @see #setExceptionListener(ExceptionListener)
    */
   public synchronized void start() throws IllegalStateException
      {
      if(_thread == null && _didstart == false)
         {
         _didstart = true;
         // perform initialization outside the run block so
         // caller gets an exception returned
         try{
            initialize();
            _successfulinit = true;
            }
         catch(Exception e)
            {
            runCleanup();
            throw new IllegalStateException(e.getMessage());
            }

         _iterations = 0;
         _thread = new Thread(getThreadGroup(), this);
         _thread.setPriority(this.getPriority());
         _thread.start();
         }
      else
         throw new IllegalStateException("Already started");

      }
	
    /**
     * Used to control whether to give up time slice to threads
	  * with the same priority.  This is significant with programs
	  * that run under "green threads" (versus native OS time-sliced
	  * threads).
     * @param yieldcount is the number of times to execute process before yielding
     * to peers.  The default is <i>DEFAULT_YIELD_COUNT</i>.  A value of 0
     * turns off yeilding.
     */  
   public void setYieldCount(int yieldcount)
	   {
	   _yieldcount = yieldcount;
	   }
	   
   public int getYieldToPeers()
       {
	   return _yieldcount;
	   }  
   /**
     * Returns the number of time the derived class's iterativeProces()
     * has been invoked successfully.
     */
   public long getIterations()
      {
      return _iterations;
      }
      
   final public void run() 
      {
      _shutdowncompleted = false;

      Thread thisThread = Thread.currentThread();

      // process until a thread altert to stop
      while(_thread == thisThread)
           {
           try{

              // In pause mode, so wait 1/2 second and loop back
              if(_paused)
                 {
                 sleepTenths(5);
                 continue;
                 }

              // Perform the application's base process
              iterativeProcess();

              // Drop thru from above was successful, so
              // increment the count
              _iterations++;

              // Test for yielding CPU cycles to another thread
              if(_yieldcount > 0)
                 {
                 // loop iterations since last yielded
                 if((_iterations % _yieldcount) == 0)
                    thisThread.yield();
                 }
                 
              } // end try block
              
           // The following exception may be due to either terminating a
           // pause or due to a stop request.  In either case, loop to the
           // top -- if due to a stop, the value the the thread variable will
           // be null, and there is logic to handle that condition.
           //
           catch(InterruptedException ignore)
             {}

           // Other exceptions are unexptected and cause the
           // thread to terminate
           catch(Exception terminate)
             {
             terminatingexception = terminate;
             _thread = null;
             // Loop back to top, where looping will terminate
             // since thread has been nulled
             }
           } // end while

      // if broke out of loop due to throwing an exception
      // (versus a stop() call) then notifiy any listeners     
      if(terminatingexception !=null)
          if(_listener != null)
              _listener.onException(terminatingexception);

      // Regardless how the loop was exited, run the cleanup
      runCleanup();

      }// end method


   private void runCleanup()
      {
      // Cleanup should be run only once
      if(_rancleanup)
         return;
         
      try{cleanup();}
      catch(Exception e){}
      finally{
             _rancleanup = true;
             _shutdowncompleted = true;
             }
      }

      
   public void setPriority(int priority){_priority = priority;}
   public int getPriority(){return _priority;}

   public int getSequenceNumber(){return _mythreadsequence;}
   public boolean isShutdownCompleted(){return _shutdowncompleted;}
   
   final public static synchronized void killAll()
      {getThreadGroup().destroy();}

   private void initialize() throws Exception
      {
      _mythreadsequence = ++threadSeqNumber;
      initializer();
      }
      
   static ThreadGroup getThreadGroup()
      {
      if(threadedgroup == null)
         threadedgroup = new ThreadGroup("Threaded");
      return threadedgroup;
      }


   public boolean isRunning()
     {return _thread != null;}

   public boolean hasTerminated()
     {return _thread == null;}
     
   /**
     * Halt execution of the thread processing
     */
   public void stop()
      {
      haltProcessing();
      }

   protected void haltProcessing()
      {
      if(_thread != null)
          {
          synchronized(_threadlock)
             {
             if(_thread != null)
                {
                // save the thread reference so it can be interrupted

                Thread o = _thread;
                // null the thread and then throw an interrupt
                _thread = null;
                o.interrupt();
                }
             }
          }
      }
      
   public synchronized void setPaused(boolean p)
      {
      _paused = p;
      if(!_paused)
        notify();
      }
      
   protected void setTerminatingException(Exception e)
      {terminatingexception = e;}
      
   public boolean isPaused()
      {return _paused;}

   public static void sleepTenths(int secs)
      {
      try{Thread.sleep(secs * 100);}
      catch(InterruptedException e){}
      }

   public static void sleepSecs(int secs)
      {
      try{Thread.sleep(secs * 1000);}
      catch(InterruptedException e){}
      }
  }

